"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatnwfilter.html
"""

from virttest.libvirt_xml import base, xcepts, accessors
from virttest.libvirt_xml.nwfilter_protocols import librarian


class NwfilterRulesProtocol(list):

    """
    List of protocol instances from classes handed out by librarian.get
    """

    @staticmethod
    def __type_check__(other):
        try:
            # Raise error if object isn't dict-like or doesn't have key
            device_tag = other['device_tag']
            # Check that we have support for this type
            librarian.get(device_tag)
        except (AttributeError, TypeError, xcepts.LibvirtXMLError):
            # Required to always raise TypeError for list API in VMXML class
            raise TypeError("Unsupported item type: %s" % str(type(other)))

    def __setitem__(self, key, value):
        self.__type_check__(value)
        super(NwfilterRulesProtocol, self).__setitem__(key, value)
        return self

    def append(self, value):
        self.__type_check__(value)
        super(NwfilterRulesProtocol, self).append(value)
        return self

    def extend(self, iterable):
        # Make sure __type_check__ happens
        for item in iterable:
            self.append(item)
        return self

    def by_device_tag(self, tag):
        result = NwfilterRulesProtocol()
        for protocol in self:
            if protocol.device_tag == tag:
                result.append(protocol)
        return result


class NwfilterXMLRules(base.LibvirtXMLBase):

    """
    Create new NwfilterXMLRules instance.

    Properties:
        rule_action: string, rule action
        rule_direction: string, rule direction
        priority: string, rule priority
        statematch: string, rule statematch
    """

    __slots__ = ('rule_action', 'rule_direction', 'rule_priority',
                 'rule_statematch')

    def __init__(self, protocol=None, virsh_instance=base.virsh):
        accessors.XMLAttribute('rule_action', self, parent_xpath='/',
                               tag_name='rule', attribute='action')
        accessors.XMLAttribute('rule_direction', self, parent_xpath='/',
                               tag_name='rule', attribute='direction')
        accessors.XMLAttribute('rule_priority', self, parent_xpath='/',
                               tag_name='rule', attribute='priority')
        accessors.XMLAttribute('rule_statematch', self, parent_xpath='/',
                               tag_name='rule', attribute='statematch')

        super(NwfilterXMLRules, self).__init__(virsh_instance=virsh_instance)
        self.xml = '<rule></rule>'

    def backup_rule(self):
        """
        Return backup rule instance

        :return: the backup of rule instance
        """
        backup = NwfilterXMLRules(virsh_instance=self.__dict_get__('virsh'))
        backup.xmltreefile = self.xmltreefile.backup_copy()

        return backup

    def get_protocol(self, protocol=None):
        """
        Return None if protocol is None, else return specific class instance

        :param protocol: specific protocol type in rules
        :return: specific protocol class instance from librarian.get
        """
        if protocol:
            protocol_class = librarian.get(protocol)
            protocol_node = self.xmltreefile.getroot().getchildren()[0]
            protocol_node.tag = protocol
            new_one = protocol_class.new_from_element(protocol_node)
            new_one.xmltreefile = self.xmltreefile
        else:
            new_one = None

        return new_one

    def new_protocol(self, **dargs):
        """
        Return a new rule protocol instance and set properties from dargs
        """
        protocol_tag = dargs.get("name")
        new_one = librarian.get(protocol_tag)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def del_protocol(self):
        """
        Delete protocol in rule xml
        """
        protocol_node = self.xmltreefile.getroot().getchildren()
        if protocol_node:
            self.xmltreefile.remove(protocol_node[0])
            self.xmltreefile.write()


class NwfilterXMLBase(base.LibvirtXMLBase):

    """
    Accessor methods for NwfilterXML class.

    Properties:
        filter_name: string, filter name
        filter_chain: string, filter name
        filter_priority: string, filter priority
        uuid: string, operates on uuid tag
        filterrefs: list, list of dictionaries describing filterref properties
    """

    __slots__ = base.LibvirtXMLBase.__slots__ + ('filter_name', 'filter_chain',
                                                 'filter_priority',
                                                 'uuid', 'filterrefs')

    __uncompareable__ = base.LibvirtXMLBase.__uncompareable__

    __schema_name__ = "nwfilter"

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLAttribute('filter_name', self, parent_xpath='/',
                               tag_name='filter', attribute='name')
        accessors.XMLAttribute('filter_chain', self, parent_xpath='/',
                               tag_name='filter', attribute='chain')
        accessors.XMLAttribute('filter_priority', self, parent_xpath='/',
                               tag_name='filter', attribute='priority')
        accessors.XMLElementText('uuid', self, parent_xpath='/',
                                 tag_name='uuid')
        accessors.XMLElementList(property_name='filterrefs',
                                 libvirtxml=self,
                                 parent_xpath='/',
                                 marshal_from=self.marshal_from_filterref,
                                 marshal_to=self.marshal_to_filterref)
        super(NwfilterXMLBase, self).__init__(virsh_instance=virsh_instance)

    @staticmethod
    def marshal_from_filterref(item, index, libvirtxml):
        """Convert a dictionary into a tag + attributes"""
        del index           # not used
        del libvirtxml      # not used
        if not isinstance(item, dict):
            raise xcepts.LibvirtXMLError("Expected a dictionary of filterref "
                                         "attributes, not a %s"
                                         % str(item))
        return ('filterref', dict(item))  # return copy of dict, not reference

    @staticmethod
    def marshal_to_filterref(tag, attr_dict, index, libvirtxml):
        """Convert a tag + attributes into a dictionary"""
        del index                    # not used
        del libvirtxml               # not used
        if tag != 'filterref':
            return None              # skip this one
        return dict(attr_dict)       # return copy of dict, not reference

    def get_rule_index(self, rule_protocol=None):
        """
        Return rule index list for specific protocol

        :param rule_protocol: the specific protocol type in rules
        :return: rule index list
        """
        rule_index = []
        source_root = self.xmltreefile.findall('rule')
        for i in range(len(source_root)):
            if rule_protocol:
                protocol_node = source_root[i].getchildren()[0]
                if protocol_node.tag == rule_protocol:
                    rule_index.append(i)
            else:
                rule_index.append(i)

        return rule_index

    def get_rule(self, rule_index=0, rule_protocol=None):
        """
        Return NwfilterXMLRules instance for specific protocol and index

        :param rule_index: rule's index number
        :param rule_protocol: the specific protocol type in rules
        :return: New initialized NwfilterXMLRules instance
        """
        index = self.get_rule_index(rule_protocol)
        if rule_index not in index:
            raise xcepts.LibvirtXMLError("rule index %s is not valid" %
                                         rule_index)
        source_root = self.xmltreefile.findall('rule')
        rulexml = NwfilterXMLRules(virsh_instance=self.__dict_get__('virsh'))
        rulexml.xmltreefile = self.xmltreefile.backup_copy()
        rulexml.xmltreefile._setroot(source_root[rule_index])
        rulexml.xmltreefile.write()
        rulexml.xmltreefile.flush()

        return rulexml

    def del_rule(self, rule_index=0):
        """
        Delete rule with specific index

        :param rule_index: rule's index number
        """
        source_root = self.xmltreefile.findall('rule')
        self.xmltreefile.remove(source_root[rule_index])
        self.xmltreefile.write()

    def set_rule(self, value, rule_index=0):
        """
        Delete rule with specific index and add new given value

        :param rule_index: rule's index number
        :param value: NwfilterXMLRules instance
        """
        if not issubclass(type(value), NwfilterXMLRules):
            raise xcepts.LibvirtXMLError(
                "Value must be a NwfilterXMLRules or subclass")
        try:
            source_root = self.xmltreefile.findall('rule')
        except KeyError as detail:
            raise xcepts.LibvirtXMLError(detail)
        if source_root[rule_index] is not None:
            self.del_rule(rule_index)
        root = self.xmltreefile.getroot()
        root.insert(rule_index, value.xmltreefile.getroot())
        self.xmltreefile.write()

    def add_rule(self, value):
        """
        Add new rule into filter

        :param value: NwfilterXMLRules instance
        """
        if not issubclass(type(value), NwfilterXMLRules):
            raise xcepts.LibvirtXMLError(
                "Value must be a NwfilterXMLRules or subclass")
        root = self.xmltreefile.getroot()
        root.append(value.xmltreefile.getroot())
        self.xmltreefile.write()

    def get_protocol_attr(self, rule_index=0, protocol=None):
        """
        Return protocol dict of specific rule index and protocol type

        :param rule_index: rule's index number
        :param protocol: the specific protocol type in rules
        :return: protocol attribute dict
        """
        rule = self.get_rule(rule_index, protocol)
        if protocol:
            protocol = rule.get_protocol(protocol)
            attr = protocol.get_attr()
        else:
            attr = None

        return attr


class NwfilterXML(NwfilterXMLBase):

    """
    Manipulators of a nwfilter through it's XML definition.
    """

    __slots__ = NwfilterXMLBase.__slots__

    def __init__(self, virsh_instance=base.virsh):
        """
        Initialize new instance with empty XML
        """
        super(NwfilterXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = u"<filter></filter>"

    @staticmethod
    def new_from_filter_dumpxml(uuid, options="", virsh_instance=base.virsh):
        """
        Return new NwfilterXML instance from virsh filter-dumpxml command

        :param uuid: filter's uuid
        :param virsh_instance: virsh module or instance to use
        :return: New initialized NwfilterXML instance
        """
        filter_xml = NwfilterXML(virsh_instance=virsh_instance)
        result = virsh_instance.nwfilter_dumpxml(uuid, options=options)
        filter_xml['xml'] = result.stdout_text.strip()

        return filter_xml

    def get_all_rules(self):
        """
        Return all rules dict with protocol attribute.

        :return: all rules dict with key as rule index number
        """
        rule_dict_attr = {}
        rule_nodes = self.xmltreefile.findall('rule')
        for i in range(len(rule_nodes)):
            if rule_nodes[i].getchildren():
                protocol_node = rule_nodes[i].getchildren()[0]
                protocol = protocol_node.tag
                pro_dict = dict(list(protocol_node.items()))
                rule_dict = dict(list(rule_nodes[i].items()))
                rule_dict.update(pro_dict)
                rule_dict['protocol'] = protocol
                rule_dict_attr[i] = rule_dict
            else:
                rule_dict = dict(list(rule_nodes[i].items()))
                rule_dict_attr[i] = rule_dict

        return rule_dict_attr

    def get_rules_dict(self, filter_name, options="",
                       virsh_instance=base.virsh):
        """
        Return all rules dict with protocol attribute for given filter

        :param filter_name: name or uuid of filter
        :param options: extra options
        :return: all rules dictionary with index as key
        """
        filxml = NwfilterXML.new_from_filter_dumpxml(filter_name,
                                                     virsh_instance=base.virsh)
        rules = filxml.get_all_rules()

        return rules

    def get_all_protocols(self, protocol=None):
        """
        Put all type of protocol into a NwfilterRulesProtocol instance.
        Return all protocols class list if protocol as None, else return
        specific protocol type class list.

        :param protocol: specific protocol type in rules
        :return: NwfilterRulesProtocol instance list
        """
        protocols = NwfilterRulesProtocol()
        all_rules = self.xmltreefile.findall('rule')

        for i in all_rules:
            protocol_node = i.getchildren()
            if protocol_node:
                if protocol:
                    # Each rule node only have one protocol node, so
                    # only use protocol_node[0]
                    if protocol_node[0].tag == protocol:
                        protocol_class = librarian.get(protocol)
                        new_one = protocol_class.new_from_element(
                            protocol_node[0])
                        protocols.device_tag = protocol
                        protocols.append(new_one)
                else:
                    protocol_class = librarian.get(protocol_node[0].tag)
                    new_one = protocol_class.new_from_element(
                        protocol_node[0])
                    protocols.device_tag = protocol_node[0].tag
                    protocols.append(new_one)

        return protocols



"""
Global background error event bus for vt test object.
This aims to share error event bus with asynchronous tasks launched by
avocado-vt, so that those tasks that have no reference to test could error the
test.

IMPORTANT: only for internal use inside avocado-vt.
"""

from six.moves import queue


class EventBus(object):
    """Event Bus."""

    def __init__(self):
        """Create the error event bus with queue.Queue."""
        self.error_events = queue.Queue()

    def put(self, event, block=True, timeout=None):
        """Put an event into the event bus."""
        self.error_events.put(event, block, timeout)

    def get(self, block=True, timeout=None):
        """Remove and return an event from the event bus."""
        try:
            return self.error_events.get(block, timeout)
        except queue.Empty:
            raise
        else:
            self.error_events.task_done()

    def __len__(self):
        """Return the event count."""
        return self.error_events.qsize()

    def get_all(self):
        """Remove and return a list of all events from the event bus."""
        error_events = []
        while len(self):
            try:
                error_events.append(self.get(block=False))
            except queue.Empty:
                pass
        return error_events

    def clear(self):
        """Clear all events in the event bus."""
        self.get_all()


error_events_bus = EventBus()



"""
Functions and classes used for logging into guests and transferring files.
"""
from __future__ import division
from six import PY3
import logging
import time
import re
import os
import shutil
import tempfile

import aexpect
from aexpect.remote import *

from avocado.core import exceptions
from avocado.utils import process

from virttest import data_dir
from virttest import utils_logfile
from virttest.remote_commander import remote_master
from virttest.remote_commander import messenger


def ssh_login_to_migrate(client, host, port, username, password, prompt, linesep="\n",
                         log_filename=None, log_function=None, timeout=10,
                         interface=None, identity_file=None,
                         status_test_command="echo $?", verbose=False, bind_ip=None,
                         preferred_authenticaton='password',
                         user_known_hosts_file='/dev/null'):
    """
    Log into a remote host (guest) using SSH (to be migrated).

    This wrapper is kept around to handle some additional SSH options like the
    preferred authentication and the known hosts file which were added while the
    PR to migrate the utility took about a year.

    :param client: The client to use ('ssh', 'telnet' or 'nc')
    :param host: Hostname or IP address
    :param port: Port to connect to
    :param username: Username (if required)
    :param password: Password (if required)
    :param prompt: Shell prompt (regular expression)
    :param linesep: The line separator to use when sending lines
            (e.g. '\\n' or '\\r\\n')
    :param log_filename: If specified, log all output to this file
    :param log_function: If specified, log all output using this function
    :param timeout: The maximal time duration (in seconds) to wait for
            each step of the login procedure (i.e. the "Are you sure" prompt
            or the password prompt)
    :param interface: The interface the neighbours attach to (only use when
                      using ipv6 linklocal address.)
    :param identity_file: Selects a file from which the identity (private key)
                          for public key authentication is read
    :param status_test_command: Command to be used for getting the last
            exit status of commands run inside the shell (used by
            cmd_status_output() and friends).
    :param bind_ip: ssh through specific interface on
                    client(specify interface ip)
    :param preferred_authenticaton: The preferred authentication of SSH connection
    :param user_known_hosts_file: one or more files to use for the user host key database
    :raise LoginError: If using ipv6 linklocal but not assign a interface that
                       the neighbour attache
    :raise LoginBadClientError: If an unknown client is requested
    :raise: Whatever handle_prompts() raises
    :return: A ShellSession object.
    """
    if client != "ssh":
        return remote_login(client, host, port, username, password, prompt,
                            linesep, log_filename, log_function, timeout,
                            interface, identity_file, status_test_command,
                            verbose, bind_ip)
    cmd = ("ssh %s -o UserKnownHostsFile=%s "
           "-o StrictHostKeyChecking=no -p %s" %
           ("-vv" if verbose else "", user_known_hosts_file, port))
    if bind_ip:
        cmd += (" -b %s" % bind_ip)
    if identity_file:
        cmd += (" -i %s" % identity_file)
    else:
        cmd += " -o PreferredAuthentications=%s" % preferred_authenticaton
    cmd += " %s@%s" % (username, host)

    if verbose:
        logging.debug("Login command: '%s'", cmd)
    session = aexpect.ShellSession(cmd, linesep=linesep, prompt=prompt,
                                   status_test_command=status_test_command)
    try:
        handle_prompts(session, username, password, prompt, timeout)
    except Exception:
        session.close()
        raise
    if log_filename:
        log_file = utils_logfile.get_log_filename(log_filename)
        session.set_output_func(utils_logfile.log_line)
        session.set_output_params((log_file,))
        session.set_log_file(os.path.basename(log_file))
    return session


def wait_for_ssh_login_to_migrate(client, host, port, username, password, prompt,
                                  linesep="\n", log_filename=None, log_function=None,
                                  timeout=240, internal_timeout=10, interface=None,
                                  preferred_authenticaton='password'):
    """
    Make multiple attempts to log into a guest until one succeeds or timeouts.

    :param timeout: Total time duration to wait for a successful login
    :param internal_timeout: The maximum time duration (in seconds) to wait for
                             each step of the login procedure (e.g. the
                             "Are you sure" prompt or the password prompt)
    :interface: The interface the neighbours attach to
                (only use when using ipv6 linklocal address.)
    :see: remote_login()
    :raise: Whatever remote_login() raises
    :return: A RemoteSession object.
    """
    logging.debug("Attempting to log into %s:%s using %s (timeout %ds)",
                  host, port, client, timeout)
    end_time = time.time() + timeout
    verbose = False
    while time.time() < end_time:
        try:
            return ssh_login_to_migrate(client, host, port, username, password, prompt,
                                        linesep, log_filename, log_function,
                                        internal_timeout, interface, verbose=verbose,
                                        preferred_authenticaton=preferred_authenticaton)
        except LoginError as error:
            logging.debug(error)
            verbose = True
        time.sleep(2)
    # Timeout expired; try one more time but don't catch exceptions
    if PY3:
        # 'preferred_authenticaton' is introduced in aexpect 1.6.1. In travis CI for python2.7,
        # it still uses aexpect 1.6.0 due to python2 compatibility problem, so E1123 will
        # be reported.
        # pylint: disable=E1123
        return remote_login(client, host, port, username, password, prompt,
                            linesep, log_filename, log_function,
                            internal_timeout, interface,
                            preferred_authenticaton=preferred_authenticaton)
    else:
        return remote_login(client, host, port, username, password, prompt,
                            linesep, log_filename, log_function,
                            internal_timeout, interface)


class AexpectIOWrapperOut(messenger.StdIOWrapperOutBase64):

    """
    Basic implementation of IOWrapper for stdout
    """

    def close(self):
        self._obj.close()

    def fileno(self):
        return os.open(self._obj, os.O_RDWR)

    def write(self, data):
        self._obj.send(data)


def remote_commander(client, host, port, username, password, prompt,
                     linesep="\n", log_filename=None, timeout=10, path=None):
    """
    Log into a remote host (guest) using SSH/Telnet/Netcat.

    :param client: The client to use ('ssh', 'telnet' or 'nc')
    :param host: Hostname or IP address
    :param port: Port to connect to
    :param username: Username (if required)
    :param password: Password (if required)
    :param prompt: Shell prompt (regular expression)
    :param linesep: The line separator to use when sending lines
            (e.g. '\\n' or '\\r\\n')
    :param log_filename: If specified, log all output to this file
    :param timeout: The maximal time duration (in seconds) to wait for
            each step of the login procedure (i.e. the "Are you sure" prompt
            or the password prompt)
    :param path: The path to place where remote_runner.py is placed.
    :raise LoginBadClientError: If an unknown client is requested
    :raise: Whatever handle_prompts() raises
    :return: A ShellSession object.
    """
    if path is None:
        path = data_dir.get_tmp_dir()
    if client == "ssh":
        cmd = ("ssh -o UserKnownHostsFile=/dev/null "
               "-o PreferredAuthentications=password "
               "-p %s %s@%s %s agent_base64" %
               (port, username, host, os.path.join(path, "remote_runner.py")))
    elif client == "telnet":
        cmd = "telnet -l %s %s %s" % (username, host, port)
    elif client == "nc":
        cmd = "nc %s %s" % (host, port)
    else:
        raise LoginBadClientError(client)

    logging.debug("Login command: '%s'", cmd)
    session = aexpect.Expect(cmd, linesep=linesep)
    try:
        handle_prompts(session, username, password, prompt, timeout)
    except Exception:
        session.close()
        raise
    if log_filename:
        log_file = utils_logfile.get_log_filename(log_filename)
        session.set_output_func(utils_logfile.log_line)
        session.set_output_params((log_file,))
        session.set_log_file(os.path.basename(log_file))

    session.send_ctrl("raw")
    # Wrap io interfaces.
    inw = messenger.StdIOWrapperInBase64(session._get_fd("tail"))
    outw = AexpectIOWrapperOut(session)
    # Create commander

    cmd = remote_master.CommanderMaster(inw, outw, False)
    return cmd


def run_remote_cmd(cmd, params, remote_runner=None, ignore_status=True):
    """
    A function to run a command on remote host.

    :param cmd: the command to be executed
    :param params: the parameter for executing
    :param remote_runner: a remote runner object on remote host
    :param ignore_status: True - not raise exception when failed
                          False - raise exception when failed

    :return: CmdResult object
    :raise: exceptions.TestFail or exceptions.TestError
    """
    try:
        if not remote_runner:
            remote_ip = params.get("server_ip", params.get("remote_ip"))
            remote_pwd = params.get("server_pwd", params.get("remote_pwd"))
            remote_user = params.get("server_user", params.get("remote_user"))
            remote_runner = RemoteRunner(host=remote_ip,
                                         username=remote_user,
                                         password=remote_pwd)

        cmdresult = remote_runner.run(cmd, ignore_status=ignore_status)
        logging.debug("Remote runner run result:\n%s", cmdresult)
        if cmdresult.exit_status and not ignore_status:
            raise exceptions.TestFail("Failed to run '%s' on remote: %s"
                                      % (cmd, cmdresult))
        return cmdresult
    except (LoginError, LoginTimeoutError,
            LoginAuthenticationError, LoginProcessTerminatedError) as e:
        logging.error(e)
        raise exceptions.TestError(e)
    except process.CmdError as cmderr:
        logging.error("Remote runner run failed:\n%s", cmderr)
        raise exceptions.TestFail("Failed to run '%s' on remote: %s"
                                  % (cmd, cmderr))


class Remote_Package(object):

    def __init__(self, address, client, username, password, port, remote_path):
        """
        Initialization of Remote Package class.

        :param address: Address of remote host(guest)
        :param client: The client to use ('ssh', 'telnet' or 'nc')
        :param username: Username (if required)
        :param password: Password (if requried)
        :param port: Port to connect to
        :param remote_path: Rmote package path
        """
        self.address = address
        self.client = client
        self.port = port
        self.username = username
        self.password = password
        self.remote_path = remote_path

        if self.client == "nc":
            self.cp_client = "rss"
            self.cp_port = 10023
        elif self.client == "ssh":
            self.cp_client = "scp"
            self.cp_port = 22
        else:
            raise LoginBadClientError(client)

    def pull_file(self, local_path, timeout=600):
        """
        Copy file from remote to local.
        """
        logging.debug("Pull remote: '%s' to local: '%s'." % (self.remote_path,
                                                             local_path))
        copy_files_from(self.address, self.cp_client, self.username,
                        self.password, self.cp_port, self.remote_path,
                        local_path, timeout=timeout)

    def push_file(self, local_path, timeout=600):
        """
        Copy file from local to remote.
        """
        logging.debug("Push local: '%s' to remote: '%s'." % (local_path,
                                                             self.remote_path))
        copy_files_to(self.address, self.cp_client, self.username,
                      self.password, self.cp_port, local_path,
                      self.remote_path, timeout=timeout)


class RemoteFile(object):

    """
    Class to handle the operations of file on remote host or guest.
    """

    def __init__(self, address, client, username, password, port,
                 remote_path, limit="", log_filename=None,
                 verbose=False, timeout=600):
        """
        Initialization of RemoteFile class.

        :param address: Address of remote host(guest)
        :param client: Type of transfer client
        :param username: Username (if required)
        :param password: Password (if requried)
        :param remote_path: Path of file which we want to edit on remote.
        :param limit: Speed limit of file transfer.
        :param log_filename: If specified, log all output to this
                             file(SCP only)
        :param verbose: If True, log some stats using logging.debug (RSS only)
        :param timeout: The time duration (in seconds) to wait for the
                        transfer tocomplete.
        """
        self.address = address
        self.client = client
        self.username = username
        self.password = password
        self.port = port
        self.remote_path = remote_path
        self.limit = limit
        self.log_filename = log_filename
        self.verbose = verbose
        self.timeout = timeout

        # Get a local_path and all actions is taken on it.
        filename = os.path.basename(self.remote_path)

        # Get a local_path.
        tmp_dir = data_dir.get_tmp_dir()
        local_file = tempfile.NamedTemporaryFile(prefix=("%s_" % filename),
                                                 dir=tmp_dir)
        self.local_path = local_file.name
        local_file.close()

        # Get a backup_path.
        backup_file = tempfile.NamedTemporaryFile(prefix=("%s_" % filename),
                                                  dir=tmp_dir)
        self.backup_path = backup_file.name
        backup_file.close()

        # Get file from remote.
        try:
            self._pull_file()
        except SCPTransferFailedError:
            # Remote file doesn't exist, create empty file on local
            self._write_local([])

        # Save a backup.
        shutil.copy(self.local_path, self.backup_path)

    def __del__(self):
        """
        Called when the instance is about to be destroyed.
        """
        self._reset_file()
        if os.path.exists(self.backup_path):
            os.remove(self.backup_path)
        if os.path.exists(self.local_path):
            os.remove(self.local_path)

    def _pull_file(self):
        """
        Copy file from remote to local.
        """
        if self.client == "test":
            shutil.copy(self.remote_path, self.local_path)
        else:
            copy_files_from(self.address, self.client, self.username,
                            self.password, self.port, self.remote_path,
                            self.local_path, self.limit, self.log_filename,
                            self.verbose, self.timeout)

    def _push_file(self):
        """
        Copy file from local to remote.
        """
        if self.client == "test":
            shutil.copy(self.local_path, self.remote_path)
        else:
            copy_files_to(self.address, self.client, self.username,
                          self.password, self.port, self.local_path,
                          self.remote_path, self.limit, self.log_filename,
                          self.verbose, self.timeout)

    def _reset_file(self):
        """
        Copy backup from local to remote.
        """
        if self.client == "test":
            shutil.copy(self.backup_path, self.remote_path)
        else:
            copy_files_to(self.address, self.client, self.username,
                          self.password, self.port, self.backup_path,
                          self.remote_path, self.limit, self.log_filename,
                          self.verbose, self.timeout)

    def _read_local(self):
        """
        Read file on local_path.

        :return: string list got from readlines().
        """
        local_file = open(self.local_path, "r")
        lines = local_file.readlines()
        local_file.close()
        return lines

    def _write_local(self, lines):
        """
        Write file on local_path. Call writelines method of File.
        """
        local_file = open(self.local_path, "w")
        local_file.writelines(lines)
        local_file.close()

    def add(self, line_list, linesep=None):
        """
        Append lines in line_list into file on remote.

        :param line_list: string consists of lines
        :param linesep: end up with a separator
        """
        lines = self._read_local()
        for line in line_list:
            lines.append("\n%s" % line)
        if linesep is not None:
            lines[-1] += linesep
        self._write_local(lines)
        self._push_file()

    def sub(self, pattern2repl_dict):
        """
        Replace the string which match the pattern
        to the value contained in pattern2repl_dict.
        """
        lines = self._read_local()
        for pattern, repl in list(pattern2repl_dict.items()):
            for index in range(len(lines)):
                line = lines[index]
                lines[index] = re.sub(pattern, repl, line)
        self._write_local(lines)
        self._push_file()

    def truncate(self, length=0):
        """
        Truncate the detail of remote file to assigned length
        Content before
        line 1
        line 2
        line 3
        remote_file.truncate(length=1)
        Content after
        line 1

        :param length: how many lines you want to keep
        """
        lines = self._read_local()
        lines = lines[0: length]
        self._write_local(lines)
        self._push_file()

    def remove(self, pattern_list):
        """
        Remove the lines in remote file which matchs a pattern
        in pattern_list.

        :param pattern_list: pattern list to be matched
        """
        lines = self._read_local()
        for line in lines:
            for pattern in pattern_list:
                if re.match(pattern, line.rstrip('\n')):
                    lines.remove(line)
                    break
        self._write_local(lines)
        self._push_file()

    def sub_else_add(self, pattern2repl_dict):
        """
        Replace the string which match the pattern.
        If no match in the all lines, append the value
        to the end of file.
        """
        lines = self._read_local()
        for pattern, repl in list(pattern2repl_dict.items()):
            no_line_match = True
            for index in range(len(lines)):
                line = lines[index]
                if re.match(pattern, line):
                    no_line_match = False
                    lines[index] = re.sub(pattern, repl, line)
            if no_line_match:
                lines.append("\n%s" % repl)
        self._write_local(lines)
        self._push_file()


class RemoteRunner(object):

    """
    Class to provide a utils.run-like method to execute command on
    remote host or guest. Provide a similar interface with utils.run
    on local.
    """

    def __init__(self, client="ssh", host=None, port="22", username="root",
                 password=None, prompt=r"[\#\$]\s*$", linesep="\n",
                 log_filename=None, timeout=240, internal_timeout=10,
                 session=None, preferred_authenticaton='password',
                 log_function=None):
        """
        Initialization of RemoteRunner. Init a session login to remote host or
        guest.

        :param client: The client to use ('ssh', 'telnet' or 'nc')
        :param host: Hostname or IP address
        :param port: Port to connect to
        :param username: Username (if required)
        :param password: Password (if required)
        :param prompt: Shell prompt (regular expression)
        :param linesep: The line separator to use when sending lines
                (e.g. '\\n' or '\\r\\n')
        :param log_filename: If specified, log all output to this file
        :param timeout: Total time duration to wait for a successful login
        :param internal_timeout: The maximal time duration (in seconds) to wait
                for each step of the login procedure (e.g. the "Are you sure"
                prompt or the password prompt)
        :param session: An existing session
        :param preferred_authenticaton: The preferred authentication of SSH connection
        :param log_function: If specified, log all output using this function
        :see: wait_for_login()
        :raise: Whatever wait_for_login() raises
        """
        if session is None:
            if host is None:
                raise exceptions.TestError(
                    "Neither host, nor session was defined!")
            self.session = wait_for_ssh_login_to_migrate(client, host, port, username,
                                                         password, prompt, linesep,
                                                         log_filename, log_function,
                                                         timeout,
                                                         internal_timeout,
                                                         preferred_authenticaton=preferred_authenticaton)
        else:
            self.session = session
        # Init stdout pipe and stderr pipe.
        self.stdout_pipe = tempfile.mktemp()
        self.stderr_pipe = tempfile.mktemp()

    def run(self, command, timeout=60, ignore_status=False):
        """
        Method to provide a utils.run-like interface to execute command on
        remote host or guest.

        :param timeout: Total time duration to wait for command return.
        :param ignore_status: If ignore_status=True, do not raise an exception,
                              no matter what the exit code of the command is.
                              Else, raise CmdError if exit code of command is
                              not zero.
        """
        # Redirect the stdout and stderr to file, Deviding error message
        # from output, and taking off the color of output. To return the same
        # result with utils.run() function.
        command = "%s 1>%s 2>%s" % (
            command, self.stdout_pipe, self.stderr_pipe)
        status, _ = self.session.cmd_status_output(command, timeout=timeout)
        output = self.session.cmd_output("cat %s;rm -f %s" %
                                         (self.stdout_pipe, self.stdout_pipe))
        errput = self.session.cmd_output("cat %s;rm -f %s" %
                                         (self.stderr_pipe, self.stderr_pipe))
        cmd_result = process.CmdResult(command=command, exit_status=status,
                                       stdout=output, stderr=errput)
        cmd_result.stdout = cmd_result.stdout_text
        cmd_result.stderr = cmd_result.stderr_text
        if status and (not ignore_status):
            raise process.CmdError(command, cmd_result)
        return cmd_result


class VMManager(object):
    """Manage VM on remote host"""

    CMD_TIMEOUT = 240

    def __init__(self, params):
        self.remote_host = params.get("server_ip")
        self.remote_user = params.get("server_user")
        self.remote_pwd = params.get("server_pwd")
        self.vm_ip = params.get("vm_ip")
        self.vm_pwd = params.get("vm_pwd")
        self.vm_user = params.get("vm_user", 'root')
        self.port = params.get("port", 22)
        if not all([self.remote_host, self.remote_user, self.remote_pwd,
                   self.vm_ip, self.vm_pwd]):
            raise exceptions.TestError("At least one of [remote_host|"
                                       "remote_user|remote_pwd|vm_ip|"
                                       "vm_pwd] is invalid!")
        self.runner = RemoteRunner(host=self.remote_host,
                                   username=self.remote_user,
                                   password=self.remote_pwd)
        self.cmd_output = self.cmd_output_safe
        self.cmd = self.cmd_output_safe

    def setup_ssh_auth(self):
        """
        Setup SSH passwordless access between remote host
        and VM, which is on the remote host.
        """
        pri_key = '~/.ssh/id_rsa'
        pub_key = '~/.ssh/id_rsa.pub'
        # Check the private key and public key file on remote host.
        cmd = "ls %s %s" % (pri_key, pub_key)
        result = self.runner.run(cmd, ignore_status=True)
        if result.exit_status:
            logging.debug("Create new SSH key pair")
            self.runner.run("ssh-keygen -t rsa -q -N '' -f %s" % pri_key)
        else:
            logging.info("SSH key pair already exist")
        session = self.runner.session
        # To avoid the host key checking
        ssh_options = "%s %s" % ("-o UserKnownHostsFile=/dev/null",
                                 "-o StrictHostKeyChecking=no")
        session.sendline("ssh-copy-id %s -i %s root@%s"
                         % (ssh_options, pub_key, self.vm_ip))

        handle_prompts(session, self.vm_user, self.vm_pwd,
                       r"[\#\$]\s*$", debug=True)

    def check_network(self, count=5, timeout=60):
        """
        Check VM network connectivity

        :param count: counter to ping
        :param timeout: seconds to wait for
        """
        logging.debug("Check VM network connectivity...")
        vm_net_connectivity = False
        sleep_time = 5
        result = ""
        cmd = "ping -c %s %s" % (count, self.vm_ip)
        while timeout > 0:
            result = self.runner.run(cmd, ignore_status=True)
            if result.exit_status:
                time.sleep(sleep_time)
                timeout -= sleep_time
                continue
            else:
                vm_net_connectivity = True
                logging.info(result.stdout_text)
                break

        if not vm_net_connectivity:
            raise exceptions.TestFail("Failed to ping %s: %s"
                                      % (self.vm_ip, result.stdout_text))

    def run_command(self, command, runner=None, ignore_status=False, timeout=CMD_TIMEOUT):
        """
        Run command in the VM.

        :param command: The command to be executed in the VM
        :param runner: The runner to execute the command
        :param ignore_status: True, not raise an exception and
            will return CmdResult object. False, raise an exception
        :param timeout: Total time to wait for command return
        :raise: exceptions.TestFail, if the command fails
        :return: CmdResult object
        """
        ssh_options = "%s %s" % ("-o UserKnownHostsFile=/dev/null",
                                 "-o StrictHostKeyChecking=no")
        cmd = 'ssh %s %s@%s "%s"' % (ssh_options, self.vm_user,
                                     self.vm_ip, command)
        ret = None
        try:
            ret = self.runner.run(cmd, timeout=timeout, ignore_status=ignore_status)
        except process.CmdError as detail:
            logging.debug("Failed to run '%s' in the VM: %s", cmd, detail)
            raise exceptions.TestFail("Failed to run '%s' in the VM: %s",
                                      cmd, detail)
        return ret

    def cmd_output_safe(self, cmd, timeout=CMD_TIMEOUT):
        """
        Unify the interface for session.cmd_output_safe()

        :param cmd: The command to execute
        :param timeout: Total time duration to wait for command return
        :return: cmd_result.stdout_text
        """
        return self.run_command(cmd, timeout=timeout).stdout_text.strip()

    def cmd_status(self, cmd, safe=True):
        """
        Unify the interface for session.cmd_status()

        :param cmd: The command to be executed
        :param safe: Ignored so far
        :return: cmd_result.exit_status
        """
        return self.run_command(cmd).exit_status

    def cmd_status_output(self, cmd, timeout=CMD_TIMEOUT):
        """
        Unify the interface for session.cmd_status_output()

        :param cmd: The command to be executed
        :param timeout: Total time duration to wait for command return
        :return: cmd_result.exit_status and cmd_result.stdout
        """
        ret = self.run_command(cmd, timeout=timeout)
        return (ret.exit_status, ret.stdout_text.strip())



"""
Intermediate module for working with XML-related virsh functions/methods.

The intention of this module is to hide the details of working with XML
from test module code.  Helper methods are all high-level and not
condusive to direct use in error-testing.  However, access to a virsh
instance is available.

All classes defined here should inherit from LibvirtXMLBase and
utilize the property-like interface provided by utils_misc.PropCanBase
to manipulate XML from the 'xml' property. Please refer to the xml_utils
module documentation for more information on working with XMLTreeFile
instances.  Please see the virsh and utils_misc modules for information
on working with Virsh and PropCanBase classes.

All properties defined in __slots__ are intended for test-module
manipulation.  External calling of accessor methods isn't forbidden,
but discouraged. Instead, test modules should use normal reference,
assignment, and delete operations on instance properties as if they
were attributes.  It's up to the test if it uses the dict-like or
instance-attribute interface.

Internally, accessor methods (get_*(), set_*(), & del_*()) should always
use __dict_get__(), __dict_set__(), and/or __dict_del__() to manipulate
properties (otherwise infinite recursion can occur).  In some cases, where
class or instance attributes are needed (ousdie of __slots__) they must
be accessed via the __super_set__(), __super_get__(), and/or __super_del__()
methods. None of the __super_*() or the __dict_*() methods are intended for use
by test-modules.

Errors originating beneath this module (e.g. w/in virsh or libvirt_vm)
should not be caught (so caller can test for them).  Errors detected
within this module should raise LibvirtXMLError or a subclass.
"""

# These are the objects considered for common use:

# all exceptions are siblings of LibvirtXMLError
from virttest.libvirt_xml.xcepts import LibvirtXMLError

from virttest.libvirt_xml.capability_xml import CapabilityXML

from virttest.libvirt_xml.network_xml import RangeList, IPXML, NetworkXML

from virttest.libvirt_xml.vm_xml import VMXML

from virttest.libvirt_xml.pool_xml import SourceXML, PoolXML

from virttest.libvirt_xml.vol_xml import VolXML

from virttest.libvirt_xml.nwfilter_xml import NwfilterXML

from virttest.libvirt_xml.sysinfo_xml import SysinfoXML

from virttest.libvirt_xml.snapshot_xml import SnapshotXML



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatdomaincaps.html
"""
import logging
import six

from virttest import xml_utils
from virttest.libvirt_xml import base, accessors, xcepts


class DomCapabilityXML(base.LibvirtXMLBase):

    """
    Handler of libvirt domcapabilities operations.

    Properties:
        features:
            DomCapFeaturesXML instance to deal with domain features
    """
    __slots__ = ('features', 'max')
    __schema_name__ = 'domcapabilities'

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementNest(property_name='features',
                                 libvirtxml=self,
                                 parent_xpath='/',
                                 tag_name='features',
                                 subclass=DomCapFeaturesXML,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLAttribute('max', self, parent_xpath='/',
                               tag_name='vcpu', attribute='max')
        super(DomCapabilityXML, self).__init__(virsh_instance)
        result = self.__dict_get__('virsh').domcapabilities()
        self['xml'] = result.stdout_text.strip()

    def get_additional_feature_list(self, cpu_mode_name):
        """
        Get additional CPU features which explicitly specified by <feature>
        tag in cpu/mode[@name='host-model'] part of virsh domcapabilities.

        In libvirt 3.9 and above, domcapabilities give the overall features supported
        by current host/qemu/kvm via CPU model and a bunch of additional features. CPU
        model is one set of features defined by libvirt in '/usr/share/libvirt/cpu_map.xml'
        Specific features described in domcapabilities are addtional features, the valid
        policy values are 'require' and 'disable', 'require' means the feature can be emulated
        by host/qemu/kvm, 'disable' which means feature is disabled by current system due to
        host CPU doesn't support or qemu/kvm has no ability to emulate it or some other
        reasons like the feature block migration and so on.

        Below is one snipped CPU host-model part output of domcapabilities.
        <feature> tag specify addtional feature policy for one certain feature type.

        virsh domcapabilities  | xmllint -xpath "/domainCapabilities/cpu/mode[@name='host-model']" -
        <mode name="host-model" supported="yes">
            <model fallback="forbid">Skylake-Client</model>
            <vendor>Intel</vendor>
            <feature policy="require" name="ss"/>
            <feature policy="require" name="hypervisor"/>
            <feature policy="require" name="tsc_adjust"/>
            <feature policy="require" name="clflushopt"/>
            <feature policy="require" name="pdpe1gb"/>
            <feature policy="require" name="invtsc"/>
        </mode>

        :param cpu_mode_name: cpu mode name, must be 'host-model' since libvirt3.9
        :return: list of features, feature is dict-like, feature name is set to dict key,
                 feature policy is set to dict value.
                 returen is like [{'ss': 'require'}, {'pdpe1gb', 'require'}]
        """
        feature_list = []  # [{feature1: policy}, {feature2: policy}, ...]
        xmltreefile = self.__dict_get__('xml')
        try:
            for mode_node in xmltreefile.findall('/cpu/mode'):
                # Get mode which name attribute is 'host-model'
                if mode_node.get('name') == cpu_mode_name:
                    for feature in mode_node.findall('feature'):
                        item = {}
                        item[feature.get('name')] = feature.get('policy')
                        # Feature invtsc doesn't support migration, so not add it to feature list
                        if 'invtsc' not in item:
                            feature_list.append(item)
        except AttributeError as elem_attr:
            logging.warn("Failed to find attribute %s" % elem_attr)
            feature_list = []
        finally:
            return feature_list

    def get_hostmodel_name(self):
        """
        Get CPU modelname which explicitly specified by <model>.text
        in cpu/mode[@name='host-model'] part of virsh domcapabilities.

        In libvirt 3.9 and above, domcapabilities give the overall features supported
        by current host/qemu/kvm via CPU model and a bunch of additional features. CPU
        model is one set of features defined by libvirt in '/usr/share/libvirt/cpu_map.xml'
        Specific features described in domcapabilities are addtional features, the valid
        policy values are 'require' and 'disable', 'require' means the feature can be emulated
        by host/qemu/kvm, 'disable' which means feature is disabled by current system due to
        host CPU doesn't support or qemu/kvm has no ability to emulate it or some other
        reasons like the feature block migration and so on.

        Below is one snipped CPU host-model part output of domcapabilities.
        <feature> tag specify addtional feature policy for one certain feature type.

        virsh domcapabilities  | xmllint -xpath "/domainCapabilities/cpu/mode[@name='host-model']" -
        <mode name="host-model" supported="yes">
            <model fallback="forbid">Skylake-Client</model>
            <vendor>Intel</vendor>
            <feature policy="require" name="ss"/>
            <feature policy="require" name="hypervisor"/>
            <feature policy="require" name="tsc_adjust"/>
            <feature policy="require" name="clflushopt"/>
            <feature policy="require" name="pdpe1gb"/>
            <feature policy="require" name="invtsc"/>
        </mode>
        Below is one snipped CPU host-model part output of domcapabilities.
        feature tag spcific policy for one specific feature type.

        virsh domcapabilities  | xmllint -xpath "/domainCapabilities/cpu/mode[@name='host-model']" -
        <mode name="host-model" supported="yes">
            <model fallback="forbid">Skylake-Client</model>
            <vendor>Intel</vendor>
            <feature policy="require" name="ss"/>
        </mode>

        :return: modelname string
        """
        xmltreefile = self.__dict_get__('xml')
        try:
            for mode_node in xmltreefile.findall('/cpu/mode'):
                if mode_node.get('name') == 'host-model':
                    return mode_node.find('model').text
        except AttributeError as elem_attr:
            logging.warn("Failed to find attribute %s" % elem_attr)
            return ''


class DomCapFeaturesXML(base.LibvirtXMLBase):
    """
    Handler of feature element in libvirt domcapabilities.

    Properties:
        gic_supported:
            string in "yes" or "no"
        gic_enums:
            list of enum dict in /gic
    """
    __slots__ = ('gic_supported', 'gic_enums')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLAttribute(property_name='gic_supported',
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='gic',
                               attribute='supported')
        accessors.AllForbidden(property_name='gic_enums',
                               libvirtxml=self)
        super(DomCapFeaturesXML, self).__init__(virsh_instance)

    def get_gic_enums(self):
        """
        Return EnumXML instance list of gic
        """
        enum_list = []
        for enum_node in self.xmltreefile.findall('/gic/enum'):
            xml_str = xml_utils.ElementTree.tostring(enum_node)
            new_enum = EnumXML()
            new_enum.xml = xml_str
            enum_list.append(new_enum)
        return enum_list


class ValueXML(base.LibvirtXMLBase):
    """
    Value elements of EnumXML
    """

    __slots__ = ('value',)

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText(property_name='value',
                                 libvirtxml=self,
                                 parent_xpath='/',
                                 tag_name='value')
        super(ValueXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = '<value/>'


class EnumXML(base.LibvirtXMLBase):
    """
    Handler of Enum element in libvirt domcapabilities

    Properties:
        name:
            string of name for enum
        values:
            list of ValueXML instance
    """
    __slots__ = ('name', 'values')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLAttribute(property_name='name',
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='enum',
                               attribute='name')
        accessors.XMLElementList(property_name='values',
                                 libvirtxml=self,
                                 parent_xpath='/',
                                 marshal_from=self.marshal_from_values,
                                 marshal_to=self.marshal_to_values)
        super(EnumXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = '<enum/>'

    @staticmethod
    def marshal_from_values(item, index, libvirtxml):
        """Convert a EnumXML instance into a tag + attributes"""
        del index           # not used
        del libvirtxml      # not used
        if isinstance(item, six.string_types):
            return ("value", {}, item)
        else:
            raise xcepts.LibvirtXMLError("Expected a str attributes,"
                                         " not a %s" % str(item))

    @staticmethod
    def marshal_to_values(tag, attr_dict, index, libvirtxml, text):
        """Convert a tag + attributes into a EnumXML instance"""
        del attr_dict                # not used
        del index                    # not used
        if tag != 'value':
            return None
        newone = ValueXML(virsh_instance=libvirtxml.virsh)
        newone.value = text
        return newone



import logging
import imp

from avocado.utils import process

from .. import propcan, xml_utils, virsh
from ..libvirt_xml import xcepts


class LibvirtXMLBase(propcan.PropCanBase):

    """
    Base class for common attributes/methods applying to all sub-classes

    Properties:
        xml:
            virtual XMLTreeFile instance
        get:
            xml filename string
        set:
            create new XMLTreeFile instance from string or filename
        del:
            deletes property, closes & unlinks any temp. files
        xmltreefile:
            XMLTreeFile instance
        virsh:
            virsh module or Virsh class instance
        set:
            validates and sets value
        get:
            returns value
        del:
            removes value
        validates:
            virtual boolean, read-only, True/False from virt-xml-validate
    """

    __slots__ = ('xml', 'virsh', 'xmltreefile', 'validates')
    __uncompareable__ = __slots__
    __schema_name__ = None

    def __init__(self, virsh_instance=virsh):
        """
        Initialize instance with connection to virsh

        :param virsh_instance: virsh module or instance to use
        """
        self.__dict_set__('xmltreefile', None)
        self.__dict_set__('validates', None)
        super(LibvirtXMLBase, self).__init__({'virsh': virsh_instance,
                                              'xml': None})
        # Can't use accessors module here, would make circular dep.

    def __str__(self):
        """
        Returns raw XML as a string
        """
        return str(self.__dict_get__('xml'))

    def __eq__(self, other):
        # Dynamic accessor methods mean we cannot compare class objects
        # directly
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        # Don't assume both instances have same comparables
        uncomparable = set(self.__uncompareable__)
        uncomparable |= set(other.__uncompareable__)
        dict_1 = {}
        dict_2 = {}
        slots = set(self.__all_slots__) | set(other.__all_slots__)
        for slot in slots - uncomparable:
            try:
                dict_1[slot] = getattr(self, slot)
            except (xcepts.LibvirtXMLNotFoundError,
                    xcepts.LibvirtXMLAccessorError):
                pass  # Unset virtual values won't have keys
            try:
                dict_2[slot] = getattr(other, slot)
            except (xcepts.LibvirtXMLNotFoundError,
                    xcepts.LibvirtXMLAccessorError):
                pass  # Unset virtual values won't have keys
        return dict_1 == dict_2

    def __contains__(self, key):
        """
        Also hide any Libvirt_xml API exceptions behind standard python behavior
        """
        try:
            return super(LibvirtXMLBase, self).__contains__(key)
        except xcepts.LibvirtXMLError:
            return False
        return True

    def set_virsh(self, value):
        """Accessor method for virsh property, make sure it's right type"""
        value_type = type(value)
        # issubclass can't work for classes using __slots__ (i.e. no __bases__)
        if hasattr(value, 'VIRSH_EXEC') or hasattr(value, 'virsh_exec'):
            self.__dict_set__('virsh', value)
        else:
            raise xcepts.LibvirtXMLError("virsh parameter must be a module "
                                         "named virsh or subclass of virsh.VirshBase "
                                         "not a %s" % str(value_type))

    def set_xml(self, value):
        """
        Accessor method for 'xml' property to load using xml_utils.XMLTreeFile
        """
        # Always check to see if a "set" accessor is being called from __init__
        if not self.__super_get__('INITIALIZED'):
            self.__dict_set__('xml', value)
        else:
            try:
                if self.__dict_get__('xml') is not None:
                    del self['xml']  # clean up old temporary files
            except KeyError:
                pass  # Allow other exceptions through
            # value could be filename or a string full of XML
            self.__dict_set__('xml', xml_utils.XMLTreeFile(value))

    def get_xml(self):
        """
        Accessor method for 'xml' property returns xmlTreeFile backup filename
        """
        return self.xmltreefile.name  # The filename

    def get_xmltreefile(self):
        """
        Return the xmltreefile object backing this instance
        """
        try:
            # don't call get_xml() recursivly
            xml = self.__dict_get__('xml')
            if xml is None:
                raise KeyError
        except (KeyError, AttributeError):
            raise xcepts.LibvirtXMLError("No xml data has been loaded")
        return xml  # XMLTreeFile loaded by set_xml() method

    def set_xmltreefile(self, value):
        """
        Point instance directly at an already initialized XMLTreeFile instance
        """
        if not issubclass(type(value), xml_utils.XMLTreeFile):
            raise xcepts.LibvirtXMLError("xmltreefile value must be XMLTreefile"
                                         " type or subclass, not a %s"
                                         % type(value))
        self.__dict_set__('xml', value)

    def del_xmltreefile(self):
        """
        Remove all backing XML
        """
        self.__dict_del__('xml')

    def copy(self):
        """
        Returns a copy of instance not sharing any references or modifications
        """
        # help keep line length short, virsh is not a property
        the_copy = self.__class__(virsh_instance=self.virsh)
        try:
            # file may not be accessible, obtain XML string value
            xmlstr = str(self.__dict_get__('xml'))
            # Create fresh/new XMLTreeFile along with tmp files from XML content
            # content
            the_copy.__dict_set__('xml', xml_utils.XMLTreeFile(xmlstr))
        except xcepts.LibvirtXMLError:  # Allow other exceptions through
            pass  # no XML was loaded yet
        return the_copy

    def get_section_string(self, xpath, index=0):
        """
        Returns the content of section in xml.

        :param xpath: xpath of xml for the section
        :param index: index of section
        """
        section = self.xmltreefile.find(xpath)
        if section is None:
            raise xcepts.LibvirtXMLNotFoundError(
                "Path %s is not found." % xpath)

        return self.xmltreefile.get_element_string(xpath, index=index)

    def get_validates(self):
        """
        Accessor method for 'validates' property returns virt-xml-validate T/F
        """
        # self.xml is the filename
        ret = self.virt_xml_validate(self.xml,
                                     self.__super_get__('__schema_name__'))
        if ret.exit_status == 0:
            return True
        else:
            logging.debug(ret)
            return False

    def set_validates(self, value):
        """
        Raises LibvirtXMLError
        """
        del value  # not needed
        raise xcepts.LibvirtXMLError("Read only property")

    def del_validates(self):
        """
        Raises LibvirtXMLError
        """
        raise xcepts.LibvirtXMLError("Read only property")

    def restore(self):
        """
        Restore current xml content to original source content
        """
        self.xmltreefile.restore()

    @staticmethod
    def virt_xml_validate(filename, schema_name=None):
        """
        Return CmdResult from running virt-xml-validate on backing XML
        """
        command = 'virt-xml-validate %s' % filename
        if schema_name:
            command += ' %s' % schema_name
        cmdresult = process.run(command, ignore_status=True)
        cmdresult.stdout = cmdresult.stdout_text
        cmdresult.stderr = cmdresult.stderr_text
        return cmdresult


def load_xml_module(path, name, type_list):
    """
    Returns named xml element's handler class

    :param path: the xml module path
    :param name: the xml module name
    :param type_list: the supported type list of xml module names
    :return: the named xml element's handler class
    """
    # Module names and tags are always all lower-case
    name = str(name).lower()
    errmsg = ("Unknown/unsupported type '%s', supported types %s"
              % (str(name), type_list))
    if name not in type_list:
        raise xcepts.LibvirtXMLError(errmsg)
    try:
        filename, pathname, description = imp.find_module(name,
                                                          [path])
        mod_obj = imp.load_module(name, filename, pathname, description)
        # Enforce capitalized class names
        return getattr(mod_obj, name.capitalize())
    except TypeError as detail:
        raise xcepts.LibvirtXMLError(errmsg + ': %s' % str(detail))
    except ImportError as detail:
        raise xcepts.LibvirtXMLError("Can't find module %s in %s: %s"
                                     % (name, path, str(detail)))
    except AttributeError as detail:
        raise xcepts.LibvirtXMLError("Can't find class %s in %s module in "
                                     "%s: %s"
                                     % (name.capitalize(), name, pathname,
                                        str(detail)))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2015
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Avocado VT plugin
"""

import os

from avocado.core.loader import loader
from avocado.core.plugin_interfaces import CLI
from avocado.utils import path as utils_path

from virttest import data_dir
from virttest import defaults
from virttest import standalone_test
from virttest.compat import get_settings_value
from virttest.standalone_test import SUPPORTED_TEST_TYPES
from virttest.standalone_test import SUPPORTED_LIBVIRT_URIS

from ..loader import VirtTestLoader


_PROVIDERS_DOWNLOAD_DIR = os.path.join(data_dir.get_test_providers_dir(),
                                       'downloads')
try:
    assert len(os.listdir(_PROVIDERS_DOWNLOAD_DIR)) != 0
except (OSError, AssertionError):
    raise ImportError("Bootstrap missing. "
                      "Execute 'avocado vt-bootstrap' or disable this "
                      "plugin to get rid of this message")


def add_basic_vt_options(parser):
    """
    Add basic vt options to parser
    """
    parser.add_argument("--vt-config", action="store", dest="vt.config",
                        help="Explicitly choose a cartesian config. When "
                        "choosing this, some options will be ignored (see "
                        "options below)")
    parser.add_argument("--vt-save-config", action="store",
                        dest="vt.save_config",
                        help="Save the resulting cartesian config to a file")
    msg = ("Choose test type (%s). Default: %%(default)s" %
           ", ".join(SUPPORTED_TEST_TYPES))
    parser.add_argument("--vt-type", action="store", dest="vt.type",
                        help=msg, default=SUPPORTED_TEST_TYPES[0])
    arch = get_settings_value('vt.common', 'arch', default=None)
    parser.add_argument("--vt-arch", help="Choose the VM architecture. "
                        "Default: %(default)s", default=arch,
                        dest='vt.common.arch')
    machine = get_settings_value('vt.common', 'machine_type',
                                 default=defaults.DEFAULT_MACHINE_TYPE)
    parser.add_argument("--vt-machine-type", help="Choose the VM machine type."
                        " Default: %(default)s", default=machine,
                        dest='vt.common.machine_type')
    parser.add_argument("--vt-guest-os", action="store",
                        dest="vt.guest_os", default=defaults.DEFAULT_GUEST_OS,
                        help="Select the guest OS to be used. If --vt-config "
                        "is provided, this will be ignored. Default: "
                        "%(default)s")
    parser.add_argument("--vt-no-filter", action="store", dest="vt.no_filter",
                        default="", help="List of space separated 'no' filters"
                        " to be passed to the config parser.  Default: "
                        "'%(default)s'")
    parser.add_argument("--vt-only-filter", action="store",
                        dest="vt.only_filter", default="", help="List of space"
                        " separated 'only' filters to be passed to the config "
                        "parser.  Default: '%(default)s'")
    parser.add_argument("--vt-filter-default-filters", nargs='+',
                        help="Allows to selectively skip certain default "
                        "filters. This uses directly 'tests-shared.cfg' and "
                        "instead of '$provider/tests.cfg' and applies "
                        "following lists of default filters, unless they are "
                        "specified as arguments: no_9p_export,no_virtio_rng,"
                        "no_pci_assignable,smallpages,default_bios,bridge,"
                        "image_backend,multihost. This can be used to eg. "
                        "run hugepages tests by filtering 'smallpages' via "
                        "this option.", dest='vt.filter.default_filters')


def add_qemu_bin_vt_option(parser):
    """
    Add qemu-bin vt option to parser
    """
    def _str_or_none(arg):
        if arg is None:
            return "Could not find one"
        else:
            return arg

    try:
        qemu_bin_path = standalone_test.find_default_qemu_paths()[0]
    except (RuntimeError, utils_path.CmdNotFoundError):
        qemu_bin_path = None
    qemu_bin = get_settings_value('vt.qemu', 'qemu_bin',
                                  default=None)
    if qemu_bin is None:    # Allow default to be None when not set in setting
        default_qemu_bin = None
        qemu_bin = qemu_bin_path
    else:
        default_qemu_bin = qemu_bin
    parser.add_argument("--vt-qemu-bin", action="store", dest="vt.qemu.qemu_bin",
                        default=default_qemu_bin, help="Path to a custom qemu"
                        " binary to be tested. If --vt-config is provided and"
                        " this flag is omitted, no attempt to set the qemu "
                        "binaries will be made. Current: %s"
                        % _str_or_none(qemu_bin))
    qemu_dst = get_settings_value('vt.qemu', 'qemu_dst_bin',
                                  default=qemu_bin_path)
    parser.add_argument("--vt-qemu-dst-bin", action="store",
                        dest="vt.qemu.qemu_dst_bin", default=qemu_dst, help="Path "
                        "to a custom qemu binary to be tested for the "
                        "destination of a migration, overrides --vt-qemu-bin. "
                        "If --vt-config is provided and this flag is omitted, "
                        "no attempt to set the qemu binaries will be made. "
                        "Current: %s" % _str_or_none(qemu_dst))


class VTRun(CLI):

    """
    Avocado VT - legacy virt-test support
    """

    name = 'vt'
    description = "Avocado VT/virt-test support to 'run' command"

    def configure(self, parser):
        """
        Add the subparser for the run action.

        :param parser: Main test runner parser.
        """
        run_subcommand_parser = parser.subcommands.choices.get('run', None)
        if run_subcommand_parser is None:
            return

        vt_compat_group_common = run_subcommand_parser.add_argument_group(
            'Virt-Test compat layer - Common options')
        vt_compat_group_qemu = run_subcommand_parser.add_argument_group(
            'Virt-Test compat layer - QEMU options')
        vt_compat_group_libvirt = run_subcommand_parser.add_argument_group(
            'Virt-Test compat layer - Libvirt options')

        add_basic_vt_options(vt_compat_group_common)
        add_qemu_bin_vt_option(vt_compat_group_qemu)
        vt_compat_group_qemu.add_argument("--vt-extra-params", nargs='*',
                                          dest="vt.extra_params",
                                          help="List of 'key=value' pairs "
                                          "passed to cartesian parser.")
        supported_uris = ", ".join(SUPPORTED_LIBVIRT_URIS)
        msg = ("Choose test connect uri for libvirt (E.g: %s). "
               "Current: %%(default)s" % supported_uris)
        uri_current = get_settings_value('vt.libvirt', 'connect_uri',
                                         default=None)
        vt_compat_group_libvirt.add_argument("--vt-connect-uri",
                                             action="store",
                                             dest="vt.libvirt.connect_uri",
                                             default=uri_current,
                                             help=msg)

    def run(self, config):
        """
        Run test modules or simple tests.

        :param config: Command line args received from the run subparser.
        """
        loader.register_plugin(VirtTestLoader)



import socket
import time
HOST = socket.gethostbyname(socket.gethostname())
s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IP)
s.bind((HOST, 0))

s.ioctl(socket.SIO_RCVALL, socket.RCVALL_ON)
time.sleep(2)
s.ioctl(socket.SIO_RCVALL, socket.RCVALL_OFF)
time.sleep(2)



"""
GlusterFS Support
This file has the functions that helps
* To create/check gluster volume.
* To start/check gluster services.
* To create gluster uri which can be used as disk image file path.
"""

import logging
import os
import re
import socket
import netaddr

from avocado.utils import process
from avocado.utils import path as utils_path
from avocado.core import exceptions

from virttest import data_dir
from virttest import utils_disk
from virttest import remote
from virttest import utils_misc
from virttest import utils_net
from virttest import error_context


class GlusterError(Exception):
    pass


class GlusterBrickError(GlusterError):

    def __init__(self, error_mgs):
        super(GlusterBrickError, self).__init__(error_mgs)
        self.error_mgs = error_mgs

    def __str__(self):
        return "Gluster: %s" % (self.error_mgs)


@error_context.context_aware
def glusterd_start():
    """
    Check for glusterd status and start it
    """
    cmd = "service glusterd status"
    output = process.run(cmd, ignore_status=True).stdout_text
    # The blank before 'active' makes a distinction with 'inactive'
    if ' active' not in output or 'running' not in output:
        cmd = "service glusterd start"
        error_context.context("Starting gluster dameon failed")
        output = process.run(cmd).stdout_text


@error_context.context_aware
def is_gluster_vol_started(vol_name, session=None):
    """
    Return true if the volume is started, if not send false

    :param vol_name: name of gluster volume
    :param session: ShellSession object of remote host
    """
    cmd = "gluster volume info %s" % vol_name
    error_context.context(
        "Gluster volume info failed for volume: %s" % vol_name)
    if session:
        vol_info = session.cmd_output(cmd)
    else:
        vol_info = process.run(cmd).stdout_text
    volume_status = re.findall(r'Status: (\S+)', vol_info)
    if 'Started' in volume_status:
        return True
    else:
        return False


@error_context.context_aware
def gluster_vol_start(vol_name, session=None):
    """
    Start the volume if it is stopped

    :param vol_name: name of gluster volume
    :param session: ShellSession object of remote host
    """
    # Check if the volume is stopped, if then start
    if not is_gluster_vol_started(vol_name, session):
        error_context.context(
            "Gluster volume start failed for volume; %s" % vol_name)
        cmd = "gluster volume start %s" % vol_name
        if session:
            session.cmd(cmd)
        else:
            process.system(cmd)
        return True
    else:
        return True


@error_context.context_aware
def gluster_vol_stop(vol_name, force=False, session=None):
    """
    Stop the volume if it is started

    :param vol_name: volume name
    :param force: Boolean for adding force option or not
    :param session: ShellSession object of remote host
    """
    # Check if the volume is started, if then stop
    if is_gluster_vol_started(vol_name, session):
        error_context.context("Gluster volume stop for volume; %s" % vol_name)
        if force:
            cmd = "echo 'y' | gluster volume stop %s force" % vol_name
        else:
            cmd = "echo 'y' | gluster volume stop %s" % vol_name
        if session:
            session.cmd(cmd)
        else:
            process.run(cmd, ignore_status=False, shell=True)
        return True
    else:
        return True


@error_context.context_aware
def gluster_vol_delete(vol_name, session=None):
    """
    Delete the volume if it is not started

    :param vol_name: name of gluster volume
    :param session: ShellSession object of remote host
    """
    # Check if the volume is stopped, if then delete
    if not is_gluster_vol_started(vol_name, session):
        error_context.context("Gluster volume delete; %s" % vol_name)
        cmd = "echo 'y' | gluster volume delete %s" % vol_name
        if session:
            session.cmd(cmd)
        else:
            process.run(cmd, ignore_status=False, shell=True)
        return True
    else:
        return False


@error_context.context_aware
def is_gluster_vol_avail(vol_name, session=None):
    """
    Check if the volume already available

    :param vol_name: name of gluster volume
    :param session: ShellSession object of remote host
    """
    cmd = "gluster volume info"
    error_context.context("Gluster volume info failed")
    if session:
        output = session.cmd_output(cmd)
    else:
        output = process.run(cmd).stdout_text
    volume_name = re.findall(r'Volume Name: (%s)\n' % vol_name, output)
    if volume_name:
        return gluster_vol_start(vol_name, session)


def gluster_brick_create(brick_path, force=False, session=None):
    """
    Creates brick

    :param brick_path: path of gluster brick
    :param force: Boolean for force delete brick or not
    :param session: ShellSession object of remote host
    """
    if session:
        cmd1_str = "session.cmd_status('[ ! -d %s ]' % brick_path)"
        cmd2_str = "session.cmd('mkdir -p %s' % brick_path)"
    else:
        cmd1_str = "os.path.isdir(brick_path)"
        cmd2_str = "os.mkdir(brick_path)"

    if eval(cmd1_str) and force:
        gluster_brick_delete(brick_path, session)
    try:
        eval(cmd2_str)
        return True
    except OSError as details:
        logging.error("Not able to create brick folder %s", details)


def gluster_brick_delete(brick_path, session=None):
    """
    Creates brick

    :param brick_path: path of gluster brick
    :param session: ShellSession object of remote host
    """
    cmd2 = 'rm -rf %s' % brick_path
    if session:
        cmd1_str = "session.cmd_status('[ ! -d %s ]' % brick_path)"
        cmd2_str = "session.cmd(cmd2)"
    else:
        cmd1_str = "os.path.isdir(brick_path)"
        cmd2_str = "process.run(cmd2, ignore_status=False, shell=True)"

    if eval(cmd1_str):
        try:
            eval(cmd2_str)
            return True
        except OSError as details:
            logging.error("Not able to delete brick folder %s", details)


@error_context.context_aware
def gluster_vol_create(vol_name, hostname, brick_path, force=False, session=None):
    """
    Gluster Volume Creation

    :param vol_name: Name of gluster volume
    :param hostname: hostname to create gluster volume
    :param force: Boolean for adding force option or not
    """
    # Create a brick
    if is_gluster_vol_avail(vol_name, session):
        gluster_vol_stop(vol_name, True, session)
        gluster_vol_delete(vol_name, session)
        gluster_brick_delete(brick_path, session)

    gluster_brick_create(brick_path, session=session)

    if force:
        force_opt = "force"
    else:
        force_opt = ""

    cmd = "gluster volume create %s %s:/%s %s" % (vol_name, hostname,
                                                  brick_path, force_opt)
    error_context.context("Volume creation failed")
    if session:
        session.cmd(cmd)
    else:
        process.system(cmd)
    return is_gluster_vol_avail(vol_name, session)


@error_context.context_aware
def glusterfs_mount(g_uri, mount_point, session=None):
    """
    Mount gluster volume to mountpoint.

    :param g_uri: stripped gluster uri from create_gluster_uri(.., True)
    :type g_uri: str
    :param mount_point: mount point, str
    :session: mount within the session if given
    """
    glusterfs_umount(g_uri, mount_point, session)
    utils_disk.mount(g_uri, mount_point, "glusterfs", None,
                     False, session)


@error_context.context_aware
def glusterfs_umount(g_uri, mount_point, session=None):
    """
    Umount gluster volume from mountpoint.

    :param g_uri: stripped gluster uri from create_gluster_uri(.., True)
    :type g_uri: str
    :param mount_point: mount point, str
    :session: mount within the session if given
    :return: if unmounted return True else return False
    """
    return utils_disk.umount(g_uri, mount_point,
                             "fuse.glusterfs", False, session)


@error_context.context_aware
def glusterfs_is_mounted(g_uri, session=None):
    """
    Check mount status of gluster volume.

    :param g_uri: stripped gluster uri from create_gluster_uri(.., True)
    :type g_uri: str
    :session: mount within the session if given
    """
    utils_disk.is_mount(g_uri, fstype="fuse.glusterfs", session=session)


@error_context.context_aware
def create_gluster_vol(params):
    vol_name = params.get("gluster_volume_name")
    force = params.get('force_recreate_gluster') == "yes"

    brick_path = params.get("gluster_brick")
    if not os.path.isabs(brick_path):  # do nothing when path is absolute
        base_dir = params.get("images_base_dir", data_dir.get_data_dir())
        brick_path = os.path.join(base_dir, brick_path)

    error_context.context("Host name lookup failed")
    hostname = socket.gethostname()
    if not hostname or hostname == "(none)":
        if_up = utils_net.get_net_if(state="UP")
        for i in if_up:
            ipv4_value = utils_net.get_net_if_addrs(i)["ipv4"]
            logging.debug("ipv4_value is %s", ipv4_value)
            if ipv4_value != []:
                ip_addr = ipv4_value[0]
                break
        hostname = ip_addr

    # Start the gluster dameon, if not started
    glusterd_start()
    # Check for the volume is already present, if not create one.
    if not is_gluster_vol_avail(vol_name) or force:
        return gluster_vol_create(vol_name, hostname, brick_path, force)
    else:
        return True


@error_context.context_aware
def create_gluster_uri(params, stripped=False):
    """
    Create gluster volume uri when using TCP:
      stripped: gluster_server:/volume
      not stripped: gluster[+transport]://gluster_server[:port]/volume
    Create gluster image uri when using unix socket:
      gluster+unix:///volume/path?socket=/path/to/socket
    """
    vol_name = params.get("gluster_volume_name")

    # Access gluster server by unix domain socket
    is_unix_socket = os.path.exists(params.get("gluster_unix_socket", ""))

    error_context.context("Host name lookup failed")
    gluster_server = params.get("gluster_server")
    if not gluster_server or is_unix_socket:
        gluster_server = socket.gethostname()
    if not gluster_server or gluster_server == "(none)":
        gluster_server = utils_net.get_host_ip_address(params)

    # Start the gluster dameon, if not started
    # Building gluster uri
    gluster_uri = None
    if stripped:
        gluster_uri = "%s:/%s" % (gluster_server, vol_name)
    else:
        volume = "/%s/" % vol_name

        if is_unix_socket:
            sock = "?socket=%s" % params["gluster_unix_socket"]
            img = params.get("image_name").split('/')[-1]
            fmt = params.get("image_format", "qcow2")
            image_path = img if params.get_boolean(
                "image_raw_device") else "%s.%s" % (img, fmt)
            gluster_uri = "gluster+unix://{v}{p}{s}".format(v=volume,
                                                            p=image_path,
                                                            s=sock)
        else:
            # Access gluster server by hostname or ip
            gluster_transport = params.get("gluster_transport")
            transport = "+%s" % gluster_transport if gluster_transport else ""

            # QEMU will send 0 which will make gluster to use the default port
            gluster_port = params.get("gluster_port")
            port = ":%s" % gluster_port if gluster_port else ""

            # [IPv6 address]
            server = "[%s]" % gluster_server if netaddr.valid_ipv6(
                                gluster_server) else gluster_server

            host = "%s%s" % (server, port)

            gluster_uri = "gluster{t}://{h}{v}".format(t=transport,
                                                       h=host,
                                                       v=volume)

    return gluster_uri


def file_exists(params, filename_path):
    sg_uri = create_gluster_uri(params, stripped=True)
    g_uri = create_gluster_uri(params, stripped=False)
    # Using directly /tmp dir because directory should be really temporary and
    # should be deleted immediately when no longer needed and
    # created directory don't file tmp dir by any data.
    tmpdir = "gmount-%s" % (utils_misc.generate_random_string(6))
    tmpdir_path = os.path.join(data_dir.get_tmp_dir(), tmpdir)
    while os.path.exists(tmpdir_path):
        tmpdir = "gmount-%s" % (utils_misc.generate_random_string(6))
        tmpdir_path = os.path.join(data_dir.get_tmp_dir(), tmpdir)
    ret = False
    try:
        try:
            os.mkdir(tmpdir_path)
            glusterfs_mount(sg_uri, tmpdir_path)
            mount_filename_path = os.path.join(tmpdir_path,
                                               filename_path[len(g_uri):])
            if os.path.exists(mount_filename_path):
                ret = True
        except Exception as e:
            logging.error("Failed to mount gluster volume %s to"
                          " mount dir %s: %s" % (sg_uri, tmpdir_path, e))
    finally:
        if glusterfs_umount(sg_uri, tmpdir_path):
            try:
                os.rmdir(tmpdir_path)
            except OSError:
                pass
        else:
            logging.warning("Unable to unmount tmp directory %s with glusterfs"
                            " mount.", tmpdir_path)
    return ret


def get_image_filename(params, image_name, image_format):
    """
    Form the image file name using gluster uri
    """
    filename = ""
    uri = create_gluster_uri(params)

    if uri.startswith("gluster+unix:"):
        filename = uri
    else:
        img_name = image_name.split('/')[-1]
        image_path = img_name if params.get_boolean(
            "image_raw_device") else "%s.%s" % (img_name, image_format)
        filename = "{vol_uri}{path}".format(vol_uri=uri, path=image_path)

    return filename


@error_context.context_aware
def gluster_allow_insecure(vol_name, session=None):
    """
    Allow gluster volume insecure

    :param vol_name: name of gluster volume
    :param session: ShellSession object of remote host
    """

    cmd1 = "gluster volume set %s server.allow-insecure on" % vol_name
    cmd2 = "gluster volume info"
    error_context.context("Volume set server.allow-insecure failed")

    if session:
        session.cmd(cmd1)
        output = session.cmd_output(cmd2)
    else:
        process.system(cmd1)
        output = process.run(cmd2).stdout_text

    match = re.findall(r'server.allow-insecure: on', output)

    if not match:
        return False
    else:
        return True


def add_rpc_insecure(filepath):
    """
    Allow glusterd RPC authority insecure
    """

    cmd = "cat %s" % filepath
    content = process.run(cmd).stdout_text
    match = re.findall(r'rpc-auth-allow-insecure on', content)
    logging.info("match is %s", match)
    if not match:
        logging.info("not match")
        cmd = "sed -i '/end-volume/i \ \ \ \ option rpc-auth-allow-insecure on' %s" % filepath
        process.system(cmd, shell=True)
        process.system("service glusterd restart; sleep 2", shell=True)


@error_context.context_aware
def gluster_nfs_disable(vol_name, session=None):
    """
    Turn-off export of volume through NFS

    :param vol_name: name of gluster volume
    :param session: Shell session for remote execution
    """

    cmd1 = "gluster volume set %s nfs.disable on" % vol_name
    cmd2 = "gluster volume info %s" % vol_name
    error_context.context("Volume set nfs.disable failed")

    if session:
        session.cmd(cmd1)
        output = session.cmd_output(cmd2)
    else:
        process.system(cmd1)
        output = process.run(cmd2).stdout_text

    return 'nfs.disable: on' in output


def setup_or_cleanup_gluster(is_setup, vol_name, brick_path="", pool_name="",
                             file_path="/etc/glusterfs/glusterd.vol", **kwargs):
    # pylint: disable=E1121
    """
    Set up or clean up glusterfs environment on localhost or remote. These actions
    can be skipped by configuration on remote in case where a static gluster instance
    is used, s. base.cfg. In this case, only the host ip is returned.

    :param is_setup: Boolean value, true for setup, false for cleanup
    :param vol_name: gluster created volume name
    :param brick_path: Dir for create glusterfs
    :param kwargs: Other parameters that need to set for gluster
    :return: ip_addr or nothing
    """
    if kwargs.get("gluster_managed_by_test", "yes") != "yes":
        if not is_setup:
            return ""
        gluster_server_ip = kwargs.get("gluster_server_ip")
        if not gluster_server_ip:
            return utils_net.get_host_ip_address()
        return gluster_server_ip

    try:
        utils_path.find_command("gluster")
    except utils_path.CmdNotFoundError:
        raise exceptions.TestSkipError("Missing command 'gluster'")
    if not brick_path:
        tmpdir = data_dir.get_tmp_dir()
        brick_path = os.path.join(tmpdir, pool_name)

    # Check gluster server apply or not
    ip_addr = kwargs.get("gluster_server_ip", "")
    session = None
    if ip_addr != "":
        remote_user = kwargs.get("gluster_server_user")
        remote_pwd = kwargs.get("gluster_server_pwd")
        remote_identity_file = kwargs.get("gluster_identity_file")
        session = remote.remote_login("ssh", ip_addr, "22",
                                      remote_user, remote_pwd, "#",
                                      identity_file=remote_identity_file)
    if is_setup:
        if ip_addr == "":
            ip_addr = utils_net.get_host_ip_address()
            add_rpc_insecure(file_path)
            glusterd_start()
            logging.debug("finish start gluster")
            logging.debug("The contents of %s: \n%s", file_path, open(file_path).read())

        gluster_vol_create(vol_name, ip_addr, brick_path, True, session)
        gluster_allow_insecure(vol_name, session)
        gluster_nfs_disable(vol_name, session)
        logging.debug("finish vol create in gluster")
        if session:
            session.close()
        return ip_addr
    else:
        gluster_vol_stop(vol_name, True, session)
        gluster_vol_delete(vol_name, session)
        gluster_brick_delete(brick_path, session)
        if session:
            session.close()
        return ""



"""
Basic nfs support for Linux host. It can support the remote
nfs mount and the local nfs set up and mount.
"""
import re
import aexpect
import logging

from avocado.utils import path
from avocado.utils import process
from avocado.utils import distro
from avocado.utils.astring import to_text
from avocado.core import exceptions

from virttest import utils_misc
from virttest import test_setup
from virttest.utils_iptables import Iptables
from virttest.utils_conn import SSHConnection
from virttest.staging import service


def nfs_exported(session=None):
    """
    Get the list for nfs file system already exported

    :param session: ShellSessionObject of remote/guest

    :return: a list of nfs that is already exported in system
    :rtype: a lit of nfs file system exported
    """
    func = process.system_output
    if session:
        func = session.cmd_output
    exportfs = to_text(func("exportfs -v"))
    if not exportfs:
        return {}

    nfs_exported_dict = {}
    for fs_info in re.findall("[/\w+]+.*?\(.*?\)", exportfs, re.S):
        fs_info = fs_info.strip().split()
        if len(fs_info) == 2:
            nfs_src = fs_info[0]
            access_ip = re.findall(r"(.*)\(", fs_info[1])[0]
            if "world" in access_ip:
                access_ip = "*"
            nfs_tag = "%s_%s" % (nfs_src, access_ip)
            permission = re.findall(r"\((.*)\)", fs_info[1])[0]
            nfs_exported_dict[nfs_tag] = permission

    return nfs_exported_dict


class Exportfs(object):

    """
    Add or remove one entry to exported nfs file system.
    """

    def __init__(self, path, client="*", options="", ori_exported=None, session=None):
        if ori_exported is None:
            ori_exported = []
        self.path = path
        self.client = client
        self.options = options.split(",")
        self.ori_exported = ori_exported
        self.entry_tag = "%s_%s" % (self.path, self.client)
        self.already_exported = False
        self.ori_options = ""
        self.session = session
        self.func = process.getoutput
        if self.session:
            self.func = self.session.cmd_output

    def is_exported(self):
        """
        Check if the directory is already exported.

        :return: If the entry is exported
        :rtype: Boolean
        """
        ori_exported = self.ori_exported or nfs_exported(session=self.session)
        if self.entry_tag in list(ori_exported.keys()):
            return True
        return False

    def need_reexport(self):
        """
        Check if the entry is already exported but the options are not
        the same as we required.

        :return: Need re export the entry or not
        :rtype: Boolean
        """
        ori_exported = self.ori_exported or nfs_exported(session=self.session)
        if self.is_exported():
            exported_options = ori_exported[self.entry_tag]
            options = [_ for _ in self.options if _ not in exported_options]
            if options:
                self.ori_options = exported_options
                return True
        return False

    def unexport(self):
        """
        Unexport an entry.
        """
        if self.is_exported():
            unexport_cmd = "exportfs -u %s:%s" % (self.client, self.path)
            self.func(unexport_cmd)
        else:
            logging.warn("Target %s %s is not exported yet."
                         "Can not unexport it." % (self.client, self.path))

    def reset_export(self):
        """
        Reset the exportfs to the original status before we export the
        specific entry.
        """
        self.unexport()
        if self.ori_options:
            tmp_options = self.options
            self.options = self.ori_options.split(",")
            self.export()
            self.options = tmp_options

    def export(self):
        """
        Export one directory if it is not in exported list.

        :return: Export nfs file system succeed or not
        """
        if self.is_exported():
            if self.need_reexport():
                self.unexport()
            else:
                self.already_exported = True
                logging.warn("Already exported target."
                             " Don't need export it again")
                return True
        export_cmd = "exportfs"
        if self.options:
            export_cmd += " -o %s" % ",".join(self.options)
        export_cmd += " %s:%s" % (self.client, self.path)
        try:
            self.func(export_cmd)
        except (process.CmdError, aexpect.ShellTimeoutError) as export_failed_err:
            logging.error("Can not export target: %s" % export_failed_err)
            return False
        return True


class Nfs(object):

    """
    Nfs class for handle nfs mount and umount. If a local nfs service is
    required, it will configure a local nfs server accroding the params.
    """

    def __init__(self, params):
        self.mount_dir = params.get("nfs_mount_dir")
        self.mount_options = params.get("nfs_mount_options")
        self.mount_src = params.get("nfs_mount_src")
        self.nfs_setup = False
        path.find_command("mount")
        self.rm_mount_dir = False
        self.rm_export_dir = False
        self.unexportfs_in_clean = False
        distro_details = distro.detect().name
        self.session = None
        self.setup_nfs_ip = params.get("nfs_server_ip", "127.0.0.1")
        self.export_dir = (params.get("export_dir") or
                           self.mount_src.split(":")[-1])
        self.export_ip = params.get("export_ip", "*")
        self.export_options = params.get("export_options", "").strip()

        if params.get("setup_remote_nfs") == "yes":
            self.nfs_setup = True
            self.setup_nfs_ip = params["nfs_server_ip"]
            nfs_server_params = {'server_ip': self.setup_nfs_ip,
                                 'server_pwd': params["nfs_server_pwd"],
                                 'server_user': params.get("nfs_server_user",
                                                           "root")}
            self.session = test_setup.remote_session(nfs_server_params)
            distro_details = utils_misc.get_distro(self.session)
            if self.session.cmd_status("exportfs -h") != 0:
                logging.error("exportfs cmd not available in remote host")

        elif params.get("setup_local_nfs") == "yes":
            self.nfs_setup = True
            self.setup_nfs_ip = "127.0.0.1"
            path.find_command("service")
            path.find_command("exportfs")

        if(params.get("setup_remote_nfs") == "yes" or
           params.get("setup_local_nfs") == "yes"):
            if 'Ubuntu' in distro_details or 'rhel' in distro_details:
                self.nfs_service = service.Service("nfs-server",
                                                   session=self.session)
            else:
                self.nfs_service = service.Service("nfs", session=self.session)

            self.rpcbind_service = service.Service("rpcbind", session=self.session)
            self.exportfs = Exportfs(self.export_dir, self.export_ip,
                                     self.export_options, session=self.session)
        self.mount_src = "%s:%s" % (self.setup_nfs_ip, self.export_dir)

    def is_mounted(self):
        """
        Check the NFS is mounted or not.

        :return: If the src is mounted as expect
        :rtype: Boolean
        """
        return utils_misc.is_mounted(self.mount_src, self.mount_dir, "nfs")

    def mount(self):
        """
        Mount source into given mount point.
        """
        return utils_misc.mount(self.mount_src, self.mount_dir, "nfs",
                                perm=self.mount_options)

    def umount(self):
        """
        Umount the given mount point.
        """
        return utils_misc.umount(self.mount_src, self.mount_dir, "nfs")

    def setup(self):
        """
        Setup NFS in host.

        Mount NFS as configured. If a local nfs is requested, setup the NFS
        service and exportfs too.
        """
        if self.nfs_setup:
            if not self.nfs_service.status():
                logging.debug("Restart NFS service.")
                self.rpcbind_service.restart()
                self.nfs_service.restart()

            if not utils_misc.check_isdir(self.export_dir, session=self.session):
                utils_misc.make_dirs(self.export_dir, session=self.session)
                self.rm_export_dir = True
            self.exportfs.export()
            self.unexportfs_in_clean = not self.exportfs.already_exported

        logging.debug("Mount %s to %s" % (self.mount_src, self.mount_dir))
        if(utils_misc.check_exists(self.mount_dir, session=self.session)
           and not utils_misc.check_isdir(self.mount_dir, session=self.session)):
            raise OSError(
                "Mount point %s is not a directory, check your setup." %
                self.mount_dir)

        if not utils_misc.check_isdir(self.mount_dir, session=self.session):
            utils_misc.make_dirs(self.mount_dir, session=self.session)
            self.rm_mount_dir = True
        self.mount()

    def cleanup(self):
        """
        Clean up the host env.

        Umount NFS from the mount point. If there has some change for exported
        file system in host when setup, also clean up that.
        """
        self.umount()
        if self.nfs_setup and self.unexportfs_in_clean:
            self.exportfs.reset_export()
            if self.rm_export_dir and utils_misc.check_isdir(self.export_dir,
                                                             session=self.session):
                utils_misc.safe_rmdir(self.export_dir)
        if self.rm_mount_dir and utils_misc.check_isdir(self.mount_dir,
                                                        session=self.session):
            utils_misc.safe_rmdir(self.mount_dir, session=self.session)


class NFSClient(object):

    """
    NFSClient class for handle nfs remotely mount and umount.
    """

    def __init__(self, params):
        self.nfs_client_ip = params.get("nfs_client_ip")
        # To Avoid host key verification failure
        ret = process.run("ssh-keygen -R %s" % self.nfs_client_ip,
                          ignore_status=True)
        stderr = ret.stderr_text
        if ret.exit_status and "No such file or directory" not in stderr:
            raise exceptions.TestFail("Failed to update host key: %s" %
                                      stderr)
        # Setup SSH connection
        self.ssh_obj = SSHConnection(params)
        ssh_timeout = int(params.get("ssh_timeout", 10))
        self.ssh_obj.conn_setup(timeout=ssh_timeout)

        self.params = params
        self.mkdir_mount_remote = False
        self.mount_dir = params.get("nfs_mount_dir")
        self.mount_options = params.get("nfs_mount_options")
        self.mount_src = params.get("nfs_mount_src")
        self.nfs_server_ip = params.get("nfs_server_ip")
        self.ssh_user = params.get("ssh_username", "root")
        self.remote_nfs_mount = params.get("remote_nfs_mount", "yes")
        self.ssh_hostkey_check = params.get("ssh_hostkey_check", "no") == "yes"
        self.ssh_cmd = "ssh %s@%s " % (self.ssh_user, self.nfs_client_ip)
        if not self.ssh_hostkey_check:
            self.ssh_cmd += "-o StrictHostKeyChecking=no "

    def is_mounted(self):
        """
        Check the NFS is mounted or not.

        :return: If the src is mounted as expect
        :rtype: Boolean
        """
        find_mountpoint_cmd = "mount | grep -E '.*%s.*%s.*'" % (self.mount_src,
                                                                self.mount_dir)
        cmd = self.ssh_cmd + "'%s'" % find_mountpoint_cmd
        logging.debug("The command: %s", cmd)
        status, output = process.getstatusoutput(cmd)
        if status:
            logging.debug("The command result: <%s:%s>", status, output)
            return False
        return True

    def setup(self):
        """
        Setup NFS client.
        """
        # Mount sharing directory to local host
        # it has been covered by class Nfs

        # Mount sharing directory to remote host
        if self.remote_nfs_mount == "yes":
            # stale file mounted causes test to fail instead
            # unmount it and perform the setup
            if self.is_mounted():
                self.umount()
            self.setup_remote()

    def umount(self):
        """
        Unmount the mount directory in remote host
        """
        logging.debug("Umount %s from %s" %
                      (self.mount_dir, self.nfs_client_ip))
        umount_cmd = self.ssh_cmd + "'umount -l %s'" % self.mount_dir
        try:
            process.system(umount_cmd, verbose=True)
        except process.CmdError:
            raise exceptions.TestFail("Failed to run: %s" % umount_cmd)

    def cleanup(self, ssh_auto_recover=False):
        """
        Cleanup NFS client.
        """
        self.umount()
        if self.mkdir_mount_remote:
            rmdir_cmd = self.ssh_cmd + "'rm -rf %s'" % self.mount_dir
            try:
                process.system(rmdir_cmd, verbose=True)
            except process.CmdError:
                raise exceptions.TestFail("Failed to run: %s" % rmdir_cmd)

        if self.is_mounted():
            raise exceptions.TestFail("Failed to umount %s" % self.mount_dir)

        # Recover SSH connection
        if ssh_auto_recover:
            self.ssh_obj.auto_recover = ssh_auto_recover
            del self.ssh_obj

    def firewall_to_permit_nfs(self):
        """
        Method to configure firewall to permit NFS to be mounted
        from remote host
        """
        # Check firewall in host permit nfs service to mount from remote server
        try:
            firewalld = service.Factory.create_service("firewalld")
            if not firewalld.status():
                firewalld.start()
            firewall_cmd = "firewall-cmd --list-all | grep services:"
            try:
                ret = process.run(firewall_cmd, shell=True)
                if not ret.exit_status:
                    firewall_services = ret.stdout_text.split(':')[1].strip().split(' ')
                    if 'nfs' not in firewall_services:
                        service_cmd = "firewall-cmd --permanent --zone=public "
                        service_cmd += "--add-service=nfs"
                        ret = process.run(service_cmd, shell=True)
                        if ret.exit_status:
                            logging.error("nfs service not added in firewall: "
                                          "%s", ret.stdout_text)
                        else:
                            logging.debug("nfs service added to firewall "
                                          "sucessfully")
                            firewalld.restart()
                    else:
                        logging.debug("nfs service already permitted by firewall")
            except process.CmdError:
                # For RHEL 6 based system firewall-cmd is not available
                logging.debug("Using iptables to permit NFS service")
                nfs_ports = []
                rule_list = []
                nfsd = service.Factory.create_service("nfs")
                rpcd = service.Factory.create_service("rpcbind")
                iptables = service.Factory.create_service("iptables")
                nfs_sysconfig = self.params.get("nfs_sysconfig_path",
                                                "/etc/sysconfig/nfs")
                tcp_port = self.params.get("nfs_tcp_port", "32803")
                udp_port = self.params.get("nfs_udp_port", "32769")
                mountd_port = self.params.get("nfs_mountd_port", "892")
                subnet_mask = self.params.get("priv_subnet", "192.168.2.0/24")
                nfs_ports.append("LOCKD_TCPPORT=%s" % tcp_port)
                nfs_ports.append("LOCKD_UDPPORT=%s" % udp_port)
                nfs_ports.append("MOUNTD_PORT=%s" % mountd_port)
                cmd_output = process.run("cat %s" % nfs_sysconfig,
                                         shell=True).stdout_text
                exist_ports = cmd_output.strip().split('\n')
                # check if the ports are already configured, if not then add it
                for each_port in nfs_ports:
                    if each_port not in exist_ports:
                        process.run("echo '%s' >> %s" %
                                    (each_port, nfs_sysconfig), shell=True)
                rpcd.restart()
                nfsd.restart()
                rule_temp = "INPUT -m state --state NEW -p %s -m multiport "
                rule_temp += "--dport 111,892,2049,%s -s %s -j ACCEPT"
                rule = rule_temp % ("tcp", tcp_port, subnet_mask)
                rule_list.append(rule)
                rule = rule_temp % ("udp", udp_port, subnet_mask)
                rule_list.append(rule)
                Iptables.setup_or_cleanup_iptables_rules(rule_list)
                iptables.restart()
        except Exception as info:
            logging.error("Firewall setting to add nfs service "
                          "failed: %s", info)

    def setup_remote(self):
        """
        Mount sharing directory to remote host.
        """
        check_mount_dir_cmd = self.ssh_cmd + "'ls -d %s'" % self.mount_dir
        logging.debug("To check if the %s exists", self.mount_dir)
        output = process.getoutput(check_mount_dir_cmd)
        if re.findall("No such file or directory", output, re.M):
            mkdir_cmd = self.ssh_cmd + "'mkdir -p %s'" % self.mount_dir
            logging.debug("Prepare to create %s", self.mount_dir)
            s, o = process.getstatusoutput(mkdir_cmd)
            if s != 0:
                raise exceptions.TestFail("Failed to run %s: %s" %
                                          (mkdir_cmd, o))
            self.mkdir_mount_remote = True

        if self.params.get("firewall_to_permit_nfs", "yes") == "yes":
            if distro.detect().name != "Ubuntu":
                self.firewall_to_permit_nfs()

        self.mount_src = "%s:%s" % (self.nfs_server_ip, self.mount_src)
        logging.debug("Mount %s to %s" % (self.mount_src, self.mount_dir))
        mount_cmd = "mount -t nfs %s %s" % (self.mount_src, self.mount_dir)
        if self.mount_options:
            mount_cmd += " -o %s" % self.mount_options
        try:
            cmd = "%s '%s'" % (self.ssh_cmd, mount_cmd)
            process.system(cmd, verbose=True)
        except process.CmdError:
            raise exceptions.TestFail("Failed to run: %s" % cmd)

        # Check if the sharing directory is mounted
        if not self.is_mounted():
            raise exceptions.TestFail("Failed to mount from %s to %s" %
                                      self.mount_src, self.mount_dir)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2015
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Avocado VT plugin
"""

import copy
import logging
import os

from avocado.core import loader
from avocado.core import output

from virttest import cartesian_config
from virttest import data_dir
from virttest import standalone_test
from virttest import storage
from virttest.compat import get_opt, set_opt

from .options import VirtTestOptionsProcess
from .test import VirtTest


LOG = logging.getLogger("avocado.app")


def guest_listing(config):
    """
    List available guest operating systems and info about image availability
    """
    if get_opt(config, 'vt.type') == 'lvsb':
        raise ValueError("No guest types available for lvsb testing")
    LOG.debug("Using %s for guest images\n",
              os.path.join(data_dir.get_data_dir(), 'images'))
    LOG.info("Available guests in config:")
    guest_name_parser = standalone_test.get_guest_name_parser(config)
    for params in guest_name_parser.get_dicts():
        base_dir = params.get("images_base_dir", data_dir.get_data_dir())
        image_name = storage.get_image_filename(params, base_dir)
        machine_type = get_opt(config, 'vt.common.machine_type')
        name = params['name'].replace('.%s' % machine_type, '')
        if os.path.isfile(image_name):
            out = name
        else:
            missing = "(missing %s)" % os.path.basename(image_name)
            out = (name + " " + output.TERM_SUPPORT.warn_header_str(missing))
        LOG.debug(out)
    LOG.debug("")


def arch_listing(config):
    """
    List available machine/archs for given guest operating systems
    """
    guest_os = get_opt(config, 'vt.guest_os')
    if guest_os is not None:
        extra = " for guest os \"%s\"" % guest_os
    else:
        extra = ""
    LOG.info("Available arch profiles%s", extra)
    guest_name_parser = standalone_test.get_guest_name_parser(config)
    machine_type = get_opt(config, 'vt.common.machine_type')
    for params in guest_name_parser.get_dicts():
        LOG.debug(params['name'].replace('.%s' % machine_type, ''))
    LOG.debug("")


class NotAvocadoVTTest(object):

    """
    Not an Avocado-vt test (for reporting purposes)
    """


class VirtTestLoader(loader.TestLoader):

    """
    Avocado loader plugin to load avocado-vt tests
    """

    name = 'vt'

    def __init__(self, config, extra_params):
        """
        Following extra_params are supported:
         * avocado_vt_extra_params: Will override the "vt_extra_params"
           of this plugins "self.config" (extends the --vt-extra-params)
        """
        vt_extra_params = extra_params.pop("avocado_vt_extra_params", None)
        super(VirtTestLoader, self).__init__(config, extra_params)
        # Avocado has renamed "args" to "config" in 84ae9a5d61, lets
        # keep making the old name available for compatibility with
        # new and old releases
        if hasattr(self, 'config'):
            self.args = self.config   # pylint: disable=E0203
        # And in case an older Avocado is used, the Loader class will
        # contain an "args" attribute instead
        else:
            self.config = self.args   # pylint: disable=E0203
        if vt_extra_params:
            # We don't want to override the original config
            self.config = copy.deepcopy(self.config)
            extra = get_opt(self.config, 'vt.extra_params')
            if extra is not None:
                extra += vt_extra_params
            else:
                extra = vt_extra_params
            set_opt(self.config, 'vt.extra_params', extra)

    def _get_parser(self):
        options_processor = VirtTestOptionsProcess(self.config)
        return options_processor.get_parser()

    def _save_parser_cartesian_config(self, parser):
        path = get_opt(self.config, 'vt.save_config')
        if path is None:
            return
        with open(path, 'w') as cartesian_config:
            cartesian_config.write("include %s\n" % parser.filename)
            for statement in (parser.only_filters + parser.no_filters +
                              parser.assignments):
                cartesian_config.write("%s\n" % statement)

    def get_extra_listing(self):
        if get_opt(self.config, 'vt_list_guests'):
            config = copy.copy(self.config)
            set_opt(config, 'vt.config', None)
            set_opt(config, 'vt.guest_os', None)
            guest_listing(config)
        if get_opt(self.config, 'vt_list_archs'):
            config = copy.copy(self.config)
            set_opt(config, 'vt.common.machine_type', None)
            set_opt(config, 'vt.common.arch', None)
            arch_listing(config)

    @staticmethod
    def get_type_label_mapping():
        """
        Get label mapping for display in test listing.

        :returns: a dictionary with the test class as key and description
                  as value.
        """
        return {VirtTest: 'VT', NotAvocadoVTTest: "!VT"}

    @staticmethod
    def get_decorator_mapping():
        """
        Get label mapping for display in test listing.

        :return: a dictionary with the test class as key and decorator
                 function as value.
        """
        term_support = output.TermSupport()
        return {VirtTest: term_support.healthy_str,
                NotAvocadoVTTest: term_support.fail_header_str}

    @staticmethod
    def _report_bad_discovery(name, reason, which_tests):
        if which_tests is loader.DiscoverMode.ALL:
            return [(NotAvocadoVTTest, {"name": "%s: %s" % (name, reason)})]
        else:
            return []

    def discover(self, url, which_tests=loader.DiscoverMode.DEFAULT):
        try:
            cartesian_parser = self._get_parser()
            self._save_parser_cartesian_config(cartesian_parser)
        except Exception as details:
            return self._report_bad_discovery(url, details, which_tests)
        if url is not None:
            try:
                cartesian_parser.only_filter(url)
            # If we have a LexerError, this means
            # the url passed is invalid in the cartesian
            # config parser, hence it should be ignored.
            # just return an empty params list and let
            # the other test plugins to handle the URL.
            except cartesian_config.ParserError as details:
                return self._report_bad_discovery(url, details, which_tests)
        elif (which_tests is loader.DiscoverMode.DEFAULT and
              not get_opt(self.config, 'vt.config')):
            # By default don't run anything unless vt.config provided
            return []
        # Create test_suite
        test_suite = []
        for params in (_ for _ in cartesian_parser.get_dicts()):
            # Evaluate the proper avocado-vt test name
            test_name = None
            if get_opt(self.config, 'vt.type') == "spice":
                short_name_map_file = params.get("_short_name_map_file")
                if "tests-variants.cfg" in short_name_map_file:
                    test_name = short_name_map_file["tests-variants.cfg"]
            if test_name is None:
                test_name = params.get("_short_name_map_file")["subtests.cfg"]
            # We want avocado to inject params coming from its multiplexer into
            # the test params. This will allow users to access avocado params
            # from inside virt tests. This feature would only work if the virt
            # test in question is executed from inside avocado.
            params['id'] = test_name
            test_parameters = {'name': test_name,
                               'vt_params': params}
            test_suite.append((VirtTest, test_parameters))
        if which_tests is loader.DiscoverMode.ALL and not test_suite:
            return self._report_bad_discovery(url, "No matching tests",
                                              which_tests)
        return test_suite



"""
    Utility module standardized on ElementTree 2.6 to minimize dependencies
    in python 2.4 systems.

    Often operations on XML files suggest making a backup copy first is a
    prudent measure.  However, it's easy to loose track of these temporary
    files and they can quickly leave a mess behind.  The TempXMLFile class
    helps by trying to clean up the temporary file whenever the instance is
    deleted, goes out of scope, or an exception is thrown.

    The XMLBackup class extends the TempXMLFile class by basing its file-
    like instances off of an automatically created TempXMLFile instead of
    pointing at the source.  Methods are provided for overwriting the backup
    copy from the source, or restoring the source from the backup.  Similar
    to TempXMLFile, the temporary backup files are automatically removed.
    Access to the original source is provided by the sourcefilename
    attribute.

    An interface for querying and manipulating XML data is provided by
    the XMLTreeFile class.  Instances of this class are BOTH file-like
    and ElementTree-like objects.  Whether or not they are constructed
    from a file or a string, the file-like instance always represents a
    temporary backup copy.  Access to the source (even when itself is
    temporary) is provided by the sourcefilename attribute, and a (closed)
    file object attribute sourcebackupfile.  See the ElementTree documentation
    for methods provided by that class.

    Finally, the TemplateXML class represents XML templates that support
    dynamic keyword substitution based on a dictionary.  Substitution keys
    in the XML template (string or file) follow the 'bash' variable reference
    style ($foo or ${bar}).  Extension of the parser is possible by subclassing
    TemplateXML and overriding the ParserClass class attribute.  The parser
    class should be an ElementTree.TreeBuilder class or subclass.  Instances
    of XMLTreeFile are returned by the parse method, which are themselves
    temporary backups of the parsed content.  See the xml_utils_unittest
    module for examples.
"""

import os
import shutil
import tempfile
import string
import logging
from six import StringIO
from xml.parsers import expat
# We *have* to use our ElementTree fork :(
from virttest import element_tree as ElementTree

# Also used by unittests
TMPPFX = 'xml_utils_temp_'
TMPSFX = '.xml'
EXSFX = '_exception_retained'
ENCODING = "UTF-8"


class TempXMLFile(object):

    """
    Temporary XML file auto-removed on instance del / module exit.
    """

    def __init__(self, suffix=TMPSFX, prefix=TMPPFX, mode="w+", buffsz=1):
        """
        Initialize temporary XML file removed on instance destruction.

        param: suffix: temporary file's suffix
        param: prefix: temporary file's prefix
        param: mode: second parameter to file()/open()
        param: buffer: third parameter to file()/open()
        """
        self.fd, self.path = tempfile.mkstemp(suffix=suffix, prefix=prefix)
        os.close(self.fd)
        self.open_file = open(self.path, mode, buffsz)
        self.name = self.open_file.name
        super(TempXMLFile, self).__init__()

    def _info(self):
        """
        Inform user that file was not auto-deleted due to exceptional exit.
        """
        logging.info("Retaining %s", self.name + EXSFX)

    def unlink(self):
        """
        Unconditionally delete file, ignoring related exceptions
        """
        try:
            os.unlink(self.name)
            self.close()
        except (OSError, IOError):
            logging.info("unlink file fail")

    def close(self):
        """
        Close file
        """
        try:
            self.open_file.close()
        except IOError:
            logging.info("close file fail")

    def seek(self, offset, whence=0):
        """
        Seek offset of one opened file
        """
        try:
            self.open_file.seek(offset, whence)
        except IOError:
            logging.info("seek file fail")

    def flush(self):
        """
        Flush buffer.
        """
        try:
            self.open_file.flush()
        except IOError:
            logging.info("flush file fail")

    def truncate(self, size):
        """
        Truncate the file to at most size bytes.
        """
        try:
            self.open_file.truncate(size)
        except IOError:
            logging.info("truncate file fail")

    def tell(self):
        """
        returns the current position of the file read/write pointer within file.
        """
        try:
            return self.open_file.tell()
        except IOError:
            logging.info("tell file fail")

    def write(self, content):
        """
        Write content to the file.
        """
        try:
            self.open_file.write(content)
        except IOError:
            logging.info("write file fail")

    def read(self, size=None):
        """
        Read file content at most size bytes, returned as a string.
        """
        try:
            if size is not None:
                return self.open_file.read(size)
            else:
                return self.open_file.read()
        except IOError:
            logging.info("read file fail")

    def readline(self, size=None):
        """
        Read next line file content from the file, as a string.
        """
        try:
            if size is not None:
                return self.open_file.readline(size)
            else:
                return self.open_file.readline()
        except IOError:
            logging.info("readline file fail")

    def readlines(self, size=None):
        """
        Read list of strings, each a line from the file.
        """
        try:
            if size is not None:
                return self.open_file.readlines(size)
            else:
                return self.open_file.readlines()
        except IOError:
            logging.info("readlines file fail")

    def __exit__(self, exc_type, exc_value, traceback):
        """
        unlink temporary backup on unexceptional module exit.
        """

        # there was an exception
        if None not in (exc_type, exc_value, traceback):
            os.rename(self.name, self.name + EXSFX)
        else:
            self.unlink()  # safe if file was renamed
        if hasattr(super(TempXMLFile, self), '__exit__'):
            super(TempXMLFile, self).__exit__(exc_type, exc_value, traceback)

    def __del__(self):
        """
        unlink temporary file on instance delete.
        """
        self.close()
        self.unlink()


class XMLBackup(TempXMLFile):

    """
    Backup file copy of XML data, automatically removed on instance destruction.
    """

    # Allow users to reference original source of XML data
    sourcefilename = None

    def __init__(self, sourcefilename):
        """
        Initialize a temporary backup from sourcefilename.
        """

        super(XMLBackup, self).__init__()
        self.sourcefilename = sourcefilename
        XMLBackup.backup(self)

    def __del__(self):
        # Drop reference, don't delete source!
        self.sourcefilename = None
        super(XMLBackup, self).__del__()

    def _info(self):
        """
        Inform user that file was not auto-deleted due to exceptional exit.
        """
        logging.info("Retaining backup of %s in %s", self.sourcefilename,
                     self.name + EXSFX)

    def backup(self):
        """
        Overwrite temporary backup with contents of original source.
        """
        super(XMLBackup, self).flush()
        super(XMLBackup, self).seek(0)
        super(XMLBackup, self).truncate(0)
        with open(self.sourcefilename, "r") as source_file:
            shutil.copyfileobj(source_file,
                               super(XMLBackup, self))
        super(XMLBackup, self).flush()

    def restore(self):
        """
        Overwrite original source with contents of temporary backup
        """
        super(XMLBackup, self).flush()
        super(XMLBackup, self).seek(0)
        with open(self.sourcefilename, "w") as source_file:
            source_file.truncate(0)
            shutil.copyfileobj(super(XMLBackup, self),
                               source_file)


class XMLTreeFile(ElementTree.ElementTree, XMLBackup):

    """
    Combination of ElementTree root and auto-cleaned XML backup file.
    """

    # Closed file object of original source or TempXMLFile
    # self.sourcefilename inherited from parent
    sourcebackupfile = None

    def __init__(self, xml):
        """
        Initialize from a string or filename containing XML source.

        param: xml: A filename or string containing XML
        """

        # xml param could be xml string or readable filename
        # If it's a string, use auto-delete TempXMLFile
        # to hold the original content.
        try:
            # Test if xml is a valid filename
            self.sourcebackupfile = open(xml, "r")
            self.sourcebackupfile.close()
            # XMLBackup init will take care of creating a copy
        except (IOError, OSError):
            # Assume xml is a string that needs a temporary source file
            self.sourcebackupfile = TempXMLFile()
            self.sourcebackupfile.write(xml)
            self.sourcebackupfile.close()
        # sourcebackupfile now safe to use for base class initialization
        XMLBackup.__init__(self, self.sourcebackupfile.name)
        try:
            ElementTree.ElementTree.__init__(self, element=None,
                                             file=self.name)
        except expat.ExpatError:
            raise IOError("Error parsing XML: '%s'" % xml)
        # Required for TemplateXML class to work
        self.write()
        self.flush()  # make sure it's on-disk

    def __str__(self):
        self.write()
        self.flush()
        xmlstr = StringIO()
        self.write(xmlstr)
        return xmlstr.getvalue()

    def backup(self):
        """Overwrite original source from current tree"""
        self.write()
        self.flush()
        # self is the 'original', so backup/restore logic is reversed
        super(XMLTreeFile, self).restore()

    def restore(self):
        """Overwrite and reparse current tree from original source"""
        # self is the 'original', so backup/restore logic is reversed
        super(XMLTreeFile, self).backup()
        try:
            ElementTree.ElementTree.__init__(self, element=None,
                                             file=self.name)
        except expat.ExpatError:
            raise IOError("Original XML is corrupt: '%s'"
                          % self.sourcebackupfile.name)

    def backup_copy(self):
        """Return a copy of instance, including copies of files"""
        return self.__class__(self.name)

    def reroot(self, xpath):
        """
        Return a copy of instance, re-rooted onto xpath
        """
        rerooted = self.backup_copy()
        element = rerooted.find(xpath)
        if element is None:
            del rerooted  # cleanup files
            raise KeyError("No element found at %s" % xpath)
        rerooted._setroot(element)
        return rerooted

    def get_parent_map(self, element=None):
        """
        Return a child to parent mapping dictionary

        param: element: Search only below this element
        """
        d = {}
        for p in self.getiterator(element):
            for c in p:
                d[c] = p
        return d

    def get_parent(self, element, relative_root=None):
        """
        Return the parent node of an element or None

        param: element: Element to retrieve parent of
        param: relative_root: Search only below this element
        """
        try:
            return self.get_parent_map(relative_root)[element]
        except KeyError:
            return None

    def get_xpath(self, element):
        """Return the XPath string formed from first-match tag names"""
        parent_map = self.get_parent_map()
        root = self.getroot()
        assert root in list(parent_map.values())
        if element == root:
            return '.'
        # List of strings reversed at end
        path_list = []
        while element != root:
            # 2.7+ ElementPath supports predicates, so:
            # element_index = list(parent_map[element]).index(element)
            # element_index += 1 # XPath indexes are 1 based
            # if element_index > 1:
            #     path_list.append(u"%s[%d]" % (element.tag, element_index))
            # else:
            #     path_list.append(u"%s" % element.tag)
            path_list.append(u"%s" % element.tag)
            element = parent_map[element]
        assert element == root
        path_list.reverse()
        return "/".join(path_list)

    def remove(self, element):
        """
        Removes a matching subelement.

        :param element: element to be removed.
        """
        self.get_parent(element).remove(element)

    def remove_by_xpath(self, xpath, remove_all=False):
        """
        Remove an element found by xpath

        :param xpath: element name or path to remove
        """
        if remove_all:
            for elem in self.findall(xpath):
                self.remove(elem)
        else:
            self.remove(self.find(xpath))  # can't remove root

    def create_by_xpath(self, xpath):
        """
        Creates all elements in simplistic xpath from root if not exist
        """
        cur_element = self.getroot()
        for tag in xpath.split('/'):
            next_element = cur_element.find(tag)
            if next_element is None:
                next_element = ElementTree.SubElement(cur_element, tag)
            cur_element = next_element

    def get_element_string(self, xpath, index=0):
        """
        Returns the string for the element on xpath.
        """
        return ElementTree.tostring(self.findall(xpath)[index])

    # This overrides the file.write() method
    def write(self, filename=None, encoding=ENCODING):
        """
        Write current XML tree to filename, or self.name if None.
        """

        if filename is None:
            filename = self.name
        # Avoid calling file.write() by mistake
        ElementTree.ElementTree.write(self, filename, encoding)

    def read(self, xml):
        self.__del__()
        self.__init__(xml)


class Sub(object):

    """String substituter using string.Template"""

    def __init__(self, **mapping):
        """Initialize substitution mapping."""

        self._mapping = mapping

    def substitute(self, text):
        """
        Use string.safe_substitute on text and return the result

        :param text: string to substitute
        """

        return string.Template(text).safe_substitute(**self._mapping)


class TemplateXMLTreeBuilder(ElementTree.XMLTreeBuilder, Sub):

    """Resolve XML templates into temporary file-backed ElementTrees"""

    BuilderClass = ElementTree.TreeBuilder

    def __init__(self, **mapping):
        """
        Initialize parser that substitutes keys with values in data

        :param mapping: values to be substituted for ${key} in XML input
        """

        Sub.__init__(self, **mapping)
        ElementTree.XMLTreeBuilder.__init__(self, target=self.BuilderClass())

    def feed(self, data):
        ElementTree.XMLTreeBuilder.feed(self, self.substitute(data))


class TemplateXML(XMLTreeFile):

    """Template-sourced XML ElementTree backed by temporary file."""

    ParserClass = TemplateXMLTreeBuilder

    def __init__(self, xml, **mapping):
        """
        Initialize from a XML string or filename, and string.template mapping.

        :param xml: A filename or string containing XML
        :param mapping: keys/values to feed with XML to string.template
        """

        self.parser = self.ParserClass(**mapping)
        # ElementTree.init calls self.parse()
        super(TemplateXML, self).__init__(xml)
        # XMLBase.__init__ calls self.write() after super init

    def parse(self, source, parser=None):
        """
        Parse source XML file or filename using TemplateXMLTreeBuilder

        :param source: XML file or filename
        :param parser: ignored
        """
        if parser is None:
            return super(TemplateXML, self).parse(source, self.parser)
        else:
            return super(TemplateXML, self).parse(source, parser)

    def restore(self):
        """
        Raise an IOError to protect the original template source.
        """

        raise IOError("Protecting template source, disallowing restore to %s" %
                      self.sourcefilename)



'''
This module implements classes that allow a user to create, enable and disable
YUM repositories on the system.
'''


import os


__all__ = ['REPO_DIR', 'YumRepo']


REPO_DIR = '/etc/yum.repos.d'


class YumRepo(object):

    '''
    Represents a YUM repository

    The goal of this class is not to give access to all features of a YUM
    Repository, but to provide a simple way to configure a valid one during
    a test run.

    Sample usage:
       >>> mainrepo = YumRepo("main", "http://download.project.org/repo",
                              "/etc/yum.repos.d/main.repo")

    Or to use a default path:
       >>> mainrepo = YumRepo("main", 'http://download.project.org/repo')

    And then:
       >>> mainrepo.save()

    When it comes to the repo URL, currently there's no support for setting a
    mirrorlist, only a baseurl.
    '''

    def __init__(self, name, baseurl, path=None):
        '''
        Initilizes a new YumRepo object

        If path is not given, it is assumed to be "$(name)s.repo" at
        the default YUM repo directory.

        :param name: the repository name
        :param path: the full path of the file that defines this repository
        '''
        self.name = name
        self.path = path
        if self.path is None:
            self.path = self._get_path_from_name(self.name)

        self.baseurl = baseurl

        self.enabled = True
        self.gpgcheck = False
        self.gpgkey = ''

    @classmethod
    def _get_path_from_name(cls, name):
        '''
        Returns the default path for the a repo of a given name

        :param name: the repository name
        :return: the default repo file path for the given name
        '''
        return os.path.join(REPO_DIR, "%s.repo" % name)

    @classmethod
    def _yum_value_for_boolean(cls, boolean):
        '''
        Returns a boolean in YUM acceptable syntax
        '''
        if boolean:
            return '1'
        else:
            return '0'

    def render(self):
        '''
        Renders the repo file

        Yes, we could use ConfigParser for this, but it produces files with
        spaces between keys and values, which look akward by YUM defaults.
        '''
        template = ("[%(name)s]\n"
                    "name=%(name)s\n"
                    "baseurl=%(baseurl)s\n"
                    "enabled=%(enabled)s\n"
                    "gpgcheck=%(gpgcheck)s\n"
                    "gpgkey=%(gpgkey)s\n")

        values = {'name': self.name,
                  'baseurl': self.baseurl,
                  'enabled': self._yum_value_for_boolean(self.enabled),
                  'gpgcheck': self._yum_value_for_boolean(self.gpgcheck),
                  'gpgkey': self.gpgkey}

        return template % values

    def save(self):
        '''
        Saves the repo file
        '''
        output_file = open(self.path, 'w')
        output_file.write(self.render())
        output_file.close()

    def remove(self):
        '''
        Removes the repo file
        '''
        if os.path.exists(self.path):
            os.unlink(self.path)



"""
tpm device support class(es)

http://libvirt.org/formatdomain.html#elementsTpm
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base


class Tpm(base.UntypedDeviceBase):

    __slots__ = ('tpm_model', 'backend')

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLAttribute('tpm_model', self,
                               parent_xpath='/',
                               tag_name='tpm',
                               attribute='model')
        accessors.XMLElementNest('backend', self, parent_xpath='/',
                                 tag_name='backend', subclass=self.Backend,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        super(Tpm, self).__init__(
            device_tag='tpm', virsh_instance=virsh_instance)

    class Backend(base.base.LibvirtXMLBase):

        """
        Tpm backend xml class.

        Properties:

        type:
            string. backend type
        version:
            string. backend version
        path:
            string. device path
        secret:
            string. encryption secret
        """
        __slots__ = ('backend_type', 'backend_version', 'device_path', 'encryption_secret')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute(property_name="backend_type",
                                   libvirtxml=self,
                                   parent_xpath='/',
                                   tag_name='backend',
                                   attribute='type')
            accessors.XMLAttribute(property_name="backend_version",
                                   libvirtxml=self,
                                   parent_xpath='/',
                                   tag_name='backend',
                                   attribute='version')
            accessors.XMLAttribute(property_name="device_path",
                                   libvirtxml=self,
                                   parent_xpath='/',
                                   tag_name='device',
                                   attribute='path')
            accessors.XMLAttribute(property_name="encryption_secret",
                                   libvirtxml=self,
                                   parent_xpath='/',
                                   tag_name='encryption',
                                   attribute='secret')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<backend/>'



"""
Module simplifying manipulation of the vm attributes described at
http://libvirt.org/formatdomain.html
"""


import logging


def set_vm_attrs(vmxml, vm_attrs):
    """
    Set element/value pairs in VMXML instance

    :param vmxml: VMXML instance of the domain
    :param vm_attrs: dict of the attribute/value pairs in VMXML
    :return the updated vmxml
    """
    for attr, value in list(vm_attrs.items()):
        logging.debug('Set %s = %s', attr, value)
        setattr(vmxml, attr, int(value) if value.isdigit() else value)
    vmxml.xmltreefile.write()
    vmxml.sync()
    return vmxml



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatnode.html
"""

import os

from virttest.libvirt_xml import base, xcepts, accessors


class CAPXML(base.LibvirtXMLBase):

    """
    The base class for capability.
    """

    def get_sysfs_sub_path(self):
        """
        return the sub path store the info of capibility.
        """
        raise NotImplementedError('get_sysfs_sub_path is not implemented.')

    @staticmethod
    def get_key2filename_dict():
        """
        Return a dict which contain the key and the name
        of info file.
        """
        raise NotImplementedError('get_key2filename_dict is not implemeneted.')

    def get_key2value_dict(self):
        """
        Reutn a dict which contain the key and the value
        in capability xml.
        """
        raise NotImplementedError('get_key2value_dict is not implemented.')


class SystemXML(CAPXML):

    """
    class for capability which type is system.
    """
    __slots__ = ('product', 'hdware_vendor', 'hdware_serial', 'hdware_uuid',
                 'firmware_vendor', 'firmversion', 'firm_release_date')

    __sysfs_sub_path__ = 'dmi/id/'

    __key2filename_dict__ = {'product': 'product_name',
                             'hdware_vendor': 'sys_vendor',
                             'hdware_serial': 'product_serial',
                             'hdware_uuid': 'product_uuid',
                             'firmware_vendor': 'bios_vendor',
                             'firmversion': 'bios_version',
                             'firm_release_date': 'bios_date'}

    @staticmethod
    def get_key2filename_dict():
        """
        Return a dict which contain the key and the name
        of info file for System node device.
        """
        return SystemXML.__key2filename_dict__

    def get_key2value_dict(self):
        """
        return the dict key2value

        key: the key in xml need to check.
        value: value in xml for this key.
        """
        key2value_dict = {}
        for key in SystemXML.__key2filename_dict__:
            key2value_dict[key] = self[key]

        return key2value_dict

    @staticmethod
    def make_sysfs_sub_path():
        """
        return __sysfs_sub_path__ immediately.
        """
        return SystemXML.__sysfs_sub_path__

    def get_sysfs_sub_path(self):
        """
        Return the sysfs_subdir.
        """
        return self.make_sysfs_sub_path()


class NetXML(CAPXML):
    """
    class for capability whose type is net.
    """
    # Example:
    #<capability type='net'>
    #<interface>eno1</interface>
    #<address>44:37:e6:95:03:e4</address>
    #<link speed='100' state='up'/>
    #<feature name='rx'/>
    #<feature name='tx'/>
    #<capability type='80203'/>
    #</capability>
    #</device>

    __slots__ = ('interface', 'address', 'link_speed', 'link_state')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText('interface', self, parent_xpath='/',
                                 tag_name='interface')
        accessors.XMLElementText('address', self, parent_xpath='/',
                                 tag_name='address')
        accessors.XMLAttribute('link_speed', self, parent_xpath='/',
                               tag_name='link', attribute='speed')
        accessors.XMLAttribute('link_state', self, parent_xpath='/',
                               tag_name='link', attribute='state')
        super(NetXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = (' <capability type=\'net\'></capability>')

    __key2filename_dict__ = {'address': 'address',
                             'link_state': 'operstate',
                             'link_speed': 'speed'}

    @staticmethod
    def get_key2filename_dict():
        """
        return the dict key2filename.
        """
        return NetXML.__key2filename_dict__

    def get_key2value_dict(self):
        """
        return the dict key2value
        """
        key2value_dict = {}
        for key in NetXML.__key2filename_dict__:
            try:
                key2value_dict[key] = self[key]
            except xcepts.LibvirtXMLNotFoundError:
                key2value_dict[key] = "-1"
        return key2value_dict


class StorageXML(CAPXML):

    """
    class for capability whose type is storage.
    """
    __slots__ = ('block', 'bus', 'driver_type')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText('block', self, parent_xpath='/',
                                 tag_name='block')
        accessors.XMLElementText('bus', self, parent_xpath='/',
                                 tag_name='bus')
        accessors.XMLElementText('driver_type', self, parent_xpath='/',
                                 tag_name='driver_type')


class PCIXML(CAPXML):

    """
    class for capability whose type is pci.
    """

    # Example:
    # <capability type='pci'>
    #   <domain>0</domain>
    #   <bus>7</bus>
    #   <slot>0</slot>
    #   <function>0</function>
    #   <product id='0x1521'>I350 Gigabit Network Connection</product>
    #   <vendor id='0x8086'>Intel Corporation</vendor>
    #   <capability type='virt_functions'>
    #     <address domain='0x0000' bus='0x08' slot='0x10' function='0x0'/>
    #     <address domain='0x0000' bus='0x08' slot='0x10' function='0x4'/>
    #   </capability>
    #   <numa node='0'/>
    # </capability>

    # or the Example of iommu:
    # <capability type="pci">
    # <domain>0</domain>
    # <bus>4</bus>
    # <slot>0</slot>
    # <function>1</function>
    # <product id="0x1639">NetXtreme II BCM5709 Gigabit Ethernet</product>
    # <vendor id="0x14e4">Broadcom Limited</vendor>
    # <iommuGroup number="15">
    #   <address bus="0x04" domain="0x0000" function="0x0" slot="0x00" />
    #   <address bus="0x04" domain="0x0000" function="0x1" slot="0x00" />
    # </iommuGroup>
    # <numa node="0" />
    # <pci-express>
    #   <link port="0" speed="5" validity="cap" width="4" />
    #   <link speed="5" validity="sta" width="4" />
    # </pci-express>
    # </capability>

    __slots__ = ('domain', 'bus', 'slot', 'function', 'product_id',
                 'vendor_id', 'virt_functions', 'numa_node',
                 'iommuGroup_number', 'iommuGroup_address',
                 'product_info')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementInt('domain', self, parent_xpath='/',
                                tag_name='domain')
        accessors.XMLElementInt('bus', self, parent_xpath='/',
                                tag_name='bus')
        accessors.XMLElementInt('slot', self, parent_xpath='/',
                                tag_name='slot')
        accessors.XMLElementInt('function', self, parent_xpath='/',
                                tag_name='function')
        accessors.XMLAttribute('product_id', self, parent_xpath='/',
                               tag_name='product', attribute='id')
        accessors.XMLAttribute('vendor_id', self, parent_xpath='/',
                               tag_name='vendor', attribute='id')
        accessors.XMLAttribute('numa_node', self, parent_xpath='/',
                               tag_name='numa', attribute='node')
        accessors.XMLAttribute('iommuGroup_number', self, parent_xpath='/',
                               tag_name='iommuGroup', attribute='number')
        accessors.XMLElementList('iommuGroup_address', self,
                                 parent_xpath='/iommuGroup',
                                 marshal_from=self.marshal_from_address,
                                 marshal_to=self.marshal_to_address)
        accessors.XMLElementList('virt_functions', self,
                                 parent_xpath='/capability',
                                 marshal_from=self.marshal_from_address,
                                 marshal_to=self.marshal_to_address)
        accessors.XMLElementText('product_info', self, parent_xpath='/',
                                 tag_name='product')
        super(PCIXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = (' <capability type=\'pci\'></capability>')

    class Address(base.LibvirtXMLBase):

        """
        Address of Virtual Function device.
        """

        # Example:
        #  <address domain='0x0000' bus='0x08' slot='0x10' function='0x0'/>
        #  <address domain='0x0000' bus='0x08' slot='0x10' function='0x4'/>

        __slots__ = ('domain', 'bus', 'slot', 'function')

        def __init__(self, virsh_instance=base.virsh):
            accessors.XMLAttribute('domain', self, parent_xpath='/',
                                   tag_name='address', attribute='domain')
            accessors.XMLAttribute('bus', self, parent_xpath='/',
                                   tag_name='address', attribute='bus')
            accessors.XMLAttribute('slot', self, parent_xpath='/',
                                   tag_name='address', attribute='slot')
            accessors.XMLAttribute('function', self, parent_xpath='/',
                                   tag_name='address', attribute='function')
            super(PCIXML.Address, self).__init__(virsh_instance=virsh_instance)
            self.xml = ('<address/>')

    @staticmethod
    def marshal_from_address(item, index, libvirtxml):
        """Convert an Address instance into tag + attributes"""
        root = item.xmltreefile.getroot()
        if root.tag == 'address':
            return (root.tag, dict(list(root.items())))
        else:
            raise xcepts.LibvirtXMLError("Expected a list of address "
                                         "instances, not a %s" % str(item))

    @staticmethod
    def marshal_to_address(tag, attr_dict, index, libvirtxml):
        """Convert a tag + attributes into an Address instance"""
        if not tag == 'address':
            return None     # Don't convert this item
        newone = PCIXML.Address(virsh_instance=libvirtxml.virsh)
        newone.update(attr_dict, excpt=xcepts.LibvirtXMLError)
        return newone

    @staticmethod
    def make_sysfs_sub_path(domain, bus, slot, function):
        """
        Make sysfs_sub_path for pci by domain,bus,slot and function.
        """
        pci_bus_path = ("%04x:%02x" % (domain, bus))
        pci_device_path = ("%04x:%02x:%02x.%01x" % (domain, bus,
                                                    slot, function))
        pci_sysfs_sub_path = ("pci_bus/%s/device/%s"
                              % (pci_bus_path, pci_device_path))

        return pci_sysfs_sub_path

    def get_sysfs_sub_path(self):
        """
        Return the sysfs_subdir in .

        Example:
            pci_bus/0000\:00/device/0000\:00\:00.0/
        """
        domain = self.domain
        bus = self.bus
        slot = self.slot
        function = self.function

        return PCIXML.make_sysfs_sub_path(domain, bus, slot, function)

    __key2filename_dict__ = {'product_id': 'device',
                             'vendor_id': 'vendor',
                             'numa_node': 'numa_node'}

    @staticmethod
    def get_key2filename_dict():
        """
        return the dict key2filename.
        key: the keys in pcixml need to check.
        filename: the name of file stored info for this key.
        """
        return PCIXML.__key2filename_dict__

    def get_key2value_dict(self):
        """
        return the dict key2value

        key: the key in xml need to check.
        value: value in xml for this key.
        """
        key2value_dict = {}
        for key in PCIXML.__key2filename_dict__:
            if key != "numa_node":
                key2value_dict[key] = self[key]
            else:
                try:
                    key2value_dict[key] = self[key]
                except xcepts.LibvirtXMLNotFoundError:
                    key2value_dict[key] = "-1"

        return key2value_dict

    def get_address_dict(self):
        """
        Return a dict contain the address.
        """
        address = {'domain': self.domain, 'bus': self.bus,
                   'slot': self.slot, 'function': self.function}
        return address


class NodedevXMLBase(base.LibvirtXMLBase):

    """
    Accessor methods for NodedevXML class.

    """

    __slots__ = ('name', 'parent', 'cap_type', 'cap',
                 'sysfs_main_path', 'host', 'fc_type',
                 'wwnn', 'wwpn', 'fabric_wwn', 'max_count', 'path')

    __schema_name__ = "nodedev"

    __sysfs_dir__ = "/sys/class"

    __type2class_dict__ = {'system': 'SystemXML',
                           'pci': 'PCIXML',
                           'usb_device': 'USBDeviceXML',
                           'usb': 'USBXML',
                           'net': 'NetXML',
                           'scsi_host': 'SCSIHostXML',
                           'scsi': 'SCSIXML',
                           'storage': 'StorageXML'}

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText('name', self, parent_xpath='/',
                                 tag_name='name')
        accessors.XMLElementText('parent', self, parent_xpath='/',
                                 tag_name='parent')
        accessors.XMLAttribute('cap_type', self, parent_xpath='/',
                               tag_name='capability', attribute='type')
        accessors.XMLElementText('host', self, parent_xpath='/capability',
                                 tag_name='host')
        accessors.XMLAttribute('fc_type', self, parent_xpath='/capability',
                               tag_name='capability', attribute='type')
        accessors.XMLAttribute('max_count', self, parent_xpath='/capability',
                               tag_name='capability', attribute='maxCount')
        accessors.XMLElementText('wwnn', self,
                                 parent_xpath='/capability/capability',
                                 tag_name='wwnn')
        accessors.XMLElementText('wwpn', self,
                                 parent_xpath='/capability/capability',
                                 tag_name='wwpn')
        accessors.XMLElementText('fabric_wwn', self,
                                 parent_xpath='/capability/capability',
                                 tag_name='fabric_wwn')
        accessors.XMLElementText('path', self, parent_xpath='/',
                                 tag_name='path')
        super(NodedevXMLBase, self).__init__(virsh_instance=virsh_instance)
        self.xml = '<device></device>'

    @staticmethod
    def get_cap_by_type(cap_type):
        """
        Init a cap class for a specific type.

        :param cap_type: the type of capability.
        :return: instanse of the cap.
        """
        cap_class_name = NodedevXMLBase.__type2class_dict__[cap_type]
        cap_class = globals()[cap_class_name]
        capxml = cap_class()

        return capxml

    def get_cap(self):
        """
        Return the capability of nodedev_xml.
        """
        try:
            cap_root = self.xmltreefile.reroot('/capability')
        except KeyError as detail:
            raise xcepts.LibvirtXMLError(detail)
        capxml = NodedevXMLBase.get_cap_by_type(self.cap_type)
        capxml.xmltreefile = cap_root
        return capxml

    def set_cap(self, value):
        """
        Set the capability by value.
        """
        if not issubclass(type(value), CAPXML):
            raise xcepts.LibvirtXMLError("value must be a CAPXML or subclass")
        # remove any existing capability block
        self.del_cap()
        root = self.xmltreefile.getroot()
        root.append(value.getroot())
        self.xmltreefile.write()

    def del_cap(self):
        """
        Delete the capability from nodedev xml.
        """
        element = self.xmltreefile.find('/capability')
        if element is not None:
            self.mltreefile.remove(element)
        self.xmltreefile.write()

    def get_sysfs_sub_path(self):
        """
        Get the sub sysfs path of the capability.
        """
        capxml = self.cap
        sysfs_sub_path = capxml.get_sysfs_sub_path()

        return sysfs_sub_path

    def get_sysfs_path(self):
        """
        Get the abs path of the capability info.
        """
        cap_type = self.cap_type
        if cap_type == 'pci':
            sysfs_main_path = self.__sysfs_dir__
            sysfs_sub_path = self.get_sysfs_sub_path()
            sysfs_path = os.path.join(sysfs_main_path, sysfs_sub_path)
        else:
            sysfs_path = self.path
        return sysfs_path


class NodedevXML(NodedevXMLBase):

    """
    class for Node device XML.
    """

    __slots__ = []

    def __init__(self, virsh_instance=base.virsh):
        """
        Initialize new instance.
        """
        super(NodedevXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = ('<device></device>')

    @staticmethod
    def new_from_dumpxml(dev_name, virsh_instance=base.virsh):
        """
        Get a instance of NodedevXML by dumpxml dev_name.
        """
        nodedevxml = NodedevXML(virsh_instance=virsh_instance)
        dumpxml_result = virsh_instance.nodedev_dumpxml(dev_name)
        if dumpxml_result.exit_status:
            stderr = dumpxml_result.stderr_text
            raise xcepts.LibvirtXMLError("Nodedev_dumpxml %s failed.\n"
                                         "Error: %s." % (dev_name, stderr))
        nodedevxml.xml = dumpxml_result.stdout_text

        return nodedevxml

    def get_key2value_dict(self):
        """
        Get the dict which contain key and value in xml.
        key: keys in nodedev xml need to check.
        value: value in xml for the key.
        """
        capxml = self.cap
        key2value_dict = capxml.get_key2value_dict()

        return key2value_dict

    def get_key2syspath_dict(self):
        """
        Get the dict which contains key and path.
        key: keys in nodedev xml need to check.
        syspath: the abs path for the file stores info for the key.
        """
        sysfs_path = self.get_sysfs_path()
        capxml = self.cap
        key2filename_dict = capxml.__class__.get_key2filename_dict()

        key2syspath_dict = {}
        for key in key2filename_dict:
            filename = key2filename_dict[key]
            abs_syspath = os.path.join(sysfs_path, filename)
            key2syspath_dict[key] = abs_syspath

        return key2syspath_dict



"""
Module simplifying manipulation of CPU model and topology part described at
http://libvirt.org/formatdomain.html
"""


import logging

from virttest.libvirt_xml import vm_xml
from virttest.utils_libvirt import libvirt_vmxml


def add_cpu_settings(vmxml, params):
    """
    Add cpu element/value pairs to the cpu xml section and keeping the existing
    contents if they are already in the xml. If the numa nodes is added, the
    number of vcpus in nodes should match the number in vcpu element. So, the
    setvm_vcpu pair need to be in params. The same for memory size(max_mem).
    The elements can be updated are list as below:
        setvm_max_mem_rt_slots = "16"
        setvm_max_mem_rt = 8192
        setvm_max_mem_rt_unit = "MiB"
        setvm_max_mem = 1536
        setvm_max_mem_unit = "MiB"
        setvm_current_mem = 1536
        setvm_current_mem_unit = "MiB"
        setvm_vcpu = 4


    :param vmxml: VMXML instance of the domain
    :param params: dict of the cpu related parameter pairs
    :return the updated vmxml
    """
    feature_list = None
    if vmxml.xmltreefile.find('cpu'):
        cpu_xml = vmxml.cpu
        feature_list = cpu_xml.get_feature_list()
    else:
        cpu_xml = vm_xml.VMCPUXML()

    if cpu_xml.xmltreefile.find('mode'):
        cpu_mode = cpu_xml.mode
    else:
        cpu_mode = params.get("cpuxml_cpu_mode", "host-model")

    if cpu_xml.xmltreefile.find("model"):
        cpu_model = cpu_xml.model
    else:
        cpu_model = params.get("cpuxml_model", "")

    if cpu_xml.xmltreefile.find("fallback"):
        cpu_model = cpu_xml.fallback
    else:
        cpu_fallback = params.get("cpuxml_fallback", "forbid")

    cpu_xml_new = vm_xml.VMCPUXML()
    cpu_xml_new.xml = "<cpu><numa/></cpu>"
    cpu_xml_new.mode = cpu_mode
    cpu_xml_new.model = cpu_model
    cpu_xml_new.fallback = cpu_fallback

    feature_name_add = params.get("cpu_feature", None)
    feature_policy_add = params.get("cpu_feature_policy", None)
    if feature_list:
        for i in range(0, len(feature_list)):
            feature_name = cpu_xml.get_feature(i).get('name')
            if feature_name == feature_name_add:
                feature_name_add = ""
            feature_policy = cpu_xml.get_feature(i).get('policy')
            cpu_xml_new.add_feature(feature_name, feature_policy)
    if feature_name_add and feature_policy_add:
        cpu_xml_new.add_feature(feature_name_add, feature_policy_add)

    if cpu_xml.xmltreefile.find('numa'):
        cpu_xml_new.numa_cell = cpu_xml.numa_cell
    else:
        cells = eval(params.get("cpuxml_numa_cell", "[]"))
        cpu_xml_new.numa_cell = vm_xml.VMCPUXML.dicts_to_cells(cells)

        # Update the vcpu and memory values to match the cell config
        # otherwise, the vm may fail to define
        vm_attrs = {k.replace('setvm_', ''): params[k] for k in params
                    if k.startswith('setvm_')}
        logging.debug(vm_attrs)
        libvirt_vmxml.set_vm_attrs(vmxml, vm_attrs)
    vmxml.cpu = cpu_xml_new
    vmxml.xmltreefile.write()
    vmxml.sync()

    return vmxml



"""
hub device support class(es)

http://libvirt.org/formatdomain.html#elementsHub
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base, librarian


class Hub(base.TypedDeviceBase):
    __slots__ = ('address',)

    def __init__(self, type_name, virsh_instance=base.base.virsh):
        super(Hub, self).__init__(device_tag='hub',
                                  type_name=type_name,
                                  virsh_instance=virsh_instance)
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={'type_name': 'usb',
                                                 'virsh_instance': virsh_instance})
    # For convenience
    Address = librarian.get('address')

    def new_hub_address(self, type_name='usb', **dargs):
        """
        Return a new hub Address instance and set properties from dargs
        """
        new_one = self.Address(type_name=type_name, virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one



import argparse

from avocado.core.settings import settings


def is_registering_settings_required():
    """Checks the characteristics of the Avocado settings API.

    And signals if the explicit registration of options is required, along
    with other API details that should be followed.

    The heuristic used here is to check for methods that are only present
    in the new API, and should be "safe enough".

    TODO: remove this once support for Avocado releases before 81.0,
    including 69.x LTS is dropped.
    """
    return all((hasattr(settings, 'add_argparser_to_option'),
                hasattr(settings, 'register_option'),
                hasattr(settings, 'as_json')))


if is_registering_settings_required():
    def get_opt(opt, name):
        """
        Compatibility handler to Avocado with configuration as dict

        :param opt: a configuration dict, usually from settings.as_dict()
        :param name: the name of the configuration key, AKA namespace
        """
        return opt.get(name)

    def set_opt(opt, name, value):
        """
        Compatibility handler to Avocado with configuration as dict

        :param opt: a configuration dict, usually from settings.as_dict()
        :param name: the name of the configuration key, AKA namespace
        :param value: the value to be set
        """
        opt[name] = value

    def set_opt_from_settings(opt, section, key, **kwargs):
        """No-op, default values are set at settings.register_option()."""
        pass

    def get_settings_value(section, key, **kwargs):
        namespace = '%s.%s' % (section, key)
        return settings.as_dict().get(namespace)
else:
    def get_opt(opt, name):
        """
        Compatibility handler for options in either argparse.Namespace or dict

        :param opt: either an argpase.Namespace instance or a dict
        :param name: the name of the attribute or key
        """
        if isinstance(opt, argparse.Namespace):
            return getattr(opt, name, None)
        else:
            return opt.get(name)

    def set_opt(opt, name, value):
        """
        Compatibility handler for options in either argparse.Namespace or dict

        :param opt: either an argpase.Namespace instance or a dict
        :param name: the name of the attribute or key
        :param value: the value to be set
        """
        if isinstance(opt, argparse.Namespace):
            setattr(opt, name, value)
        else:
            opt[name] = value

    def set_opt_from_settings(opt, section, key, **kwargs):
        """Sets option default value from the configuration file."""
        value = settings.get_value(section, key, **kwargs)
        namespace = '%s.%s' % (section, key)
        set_opt(opt, namespace, value)

    def get_settings_value(section, key, **kwargs):
        return settings.get_value(section, key, **kwargs)
